# Toshiba MeP-c4 - miasm architecture definition
# Guillaume Valadon <guillaume@valadon.net>

from builtins import range
from miasm.core.cpu import *
from miasm.core.utils import Disasm_Exception
from miasm.expression.expression import ExprId, ExprInt, ExprLoc, \
    ExprMem, ExprOp, is_expr
from miasm.core.asm_ast import AstId, AstMem

from miasm.arch.mep.regs import *
import miasm.arch.mep.regs as mep_regs_module  # will be used to set mn_mep.regs
from miasm.ir.ir import color_expr_html


# Note: pyparsing is used to alter the way special operands are parsed
from pyparsing import Literal, Group, Word, hexnums


# These definitions will help parsing dereferencing instructions (i.e. that uses
# parenthesis) with pyparsing
LPARENTHESIS = Literal("(")
RPARENTHESIS = Literal(")")
PLUSSIGN = Literal("+")
HEX_INTEGER = str_int_pos | str_int_neg


def ExprInt2SignedString(expr, pos_fmt="%d", neg_fmt="%d", size=None, offset=0):
    """Return the signed string corresponding to an ExprInt

       Note: this function is only useful to mimic objdump output"""

    # Apply a mask to the integer
    if size is None:
        mask_length = expr.size
    else:
        mask_length = size
    mask = (1 << mask_length) - 1
    value = int(expr) & mask

    # Return a signed integer if necessary
    if (value >> mask_length - 1) == 1:
        value = offset - ((value ^ mask) + 1)
        if value < 0:
            return "-" + neg_fmt % -value
    else:
        value += offset

    return pos_fmt % value


class instruction_mep(instruction):
    """Generic MeP-c4 instruction

    Notes:
        - this object is used to build internal miasm instructions based
          on mnemonics
        - it must be implemented !
    """

    @staticmethod
    def arg2str(expr, pos=None, loc_db=None):
        """Convert mnemonics arguments into readable strings according to the
        MeP-c4 architecture manual and their internal types

        Notes:
            - it must be implemented ! However, a simple 'return str(expr)'
              could do the trick.
            - it is used to mimic objdump output

        Args:
            expr: argument as a miasm expression
            pos: position index in the arguments list
        """

        if isinstance(expr, ExprId) or isinstance(expr, ExprInt):
            return str(expr)

        elif isinstance(expr, ExprLoc):
            if loc_db is not None:
                return loc_db.pretty_str(expr.loc_key)
            else:
                return str(expr)

        elif isinstance(expr, ExprMem) and (isinstance(expr.ptr, ExprId) or isinstance(expr.ptr, ExprInt)):
            return "(%s)" % expr.ptr

        elif isinstance(expr, ExprMem) and isinstance(expr.ptr, ExprOp):
            return "0x%X(%s)" % (int(expr.ptr.args[1]), expr.ptr.args[0])

        # Raise an exception if the expression type was not processed
        message = "instruction_mep.arg2str(): don't know what \
                   to do with a '%s' instance." % type(expr)
        raise Disasm_Exception(message)

    @staticmethod
    def arg2html(expr, pos=None, loc_db=None):
        """Convert mnemonics arguments into readable html strings according to the
        MeP-c4 architecture manual and their internal types

        Notes:
            - it must be implemented ! However, a simple 'return str(expr)'
              could do the trick.
            - it is used to mimic objdump output

        Args:
            expr: argument as a miasm expression
            pos: position index in the arguments list
        """

        if isinstance(expr, ExprId) or isinstance(expr, ExprInt) or isinstance(expr, ExprLoc):
            return color_expr_html(expr, loc_db)

        elif isinstance(expr, ExprMem) and (isinstance(expr.ptr, ExprId) or isinstance(expr.ptr, ExprInt)):
            return "(%s)" % color_expr_html(expr.ptr, loc_db)

        elif isinstance(expr, ExprMem) and isinstance(expr.ptr, ExprOp):
            return "%s(%s)" % (
                color_expr_html(expr.ptr.args[1], loc_db),
                color_expr_html(expr.ptr.args[0], loc_db)
            )

        # Raise an exception if the expression type was not processed
        message = "instruction_mep.arg2str(): don't know what \
                   to do with a '%s' instance." % type(expr)
        raise Disasm_Exception(message)

    def __str__(self):
        """Return the mnemonic as a string.

        Note:
            - it is not mandatory as the instruction class already implement
              it. It used to get rid of the padding between the opcode and the
              arguments.
            - most of this code is copied from miasm/core/cpu.py
        """

        o = "%s" % self.name

        if self.name == "SSARB":
            # The first operand is displayed in decimal, not in hex
            o += " %d" % int(self.args[0])
            o += self.arg2str(self.args[1])

        elif self.name in ["MOV", "ADD"] and isinstance(self.args[1], ExprInt):
            # The second operand is displayed in decimal, not in hex
            o += " " + self.arg2str(self.args[0])
            o += ", %s" % ExprInt2SignedString(self.args[1])

        elif "CPI" in self.name:
            # The second operand ends with the '+' sign
            o += " " + self.arg2str(self.args[0])
            deref_reg_str = self.arg2str(self.args[1])
            o += ", %s+)" % deref_reg_str[:-1]  # GV: looks ugly

        elif self.name[0] in ["S", "L"] and self.name[-3:] in ["CPA", "PM0", "PM1"]:
            # The second operand ends with the '+' sign
            o += " " + self.arg2str(self.args[0])
            deref_reg_str = self.arg2str(self.args[1])
            o += ", %s+)" % deref_reg_str[:-1]  # GV: looks ugly
            # The third operand is displayed in decimal, not in hex
            o += ", %s" % ExprInt2SignedString(self.args[2])

        elif len(self.args) == 2 and self.name in ["SB", "SH", "LBU", "LB", "LH", "LW"] and \
                isinstance(self.args[1], ExprMem) and isinstance(self.args[1].ptr, ExprOp):  # Major Opcodes #12
            # The second operand is an offset to a register
            o += " " + self.arg2str(self.args[0])
            o += ", %s" % ExprInt2SignedString(self.args[1].ptr.args[1], "0x%X")
            o += "(%s)" % self.arg2str(self.args[1].ptr.args[0])

        elif len(self.args) == 2 and self.name in ["SWCP", "LWCP", "SMCP", "LMCP"] \
                and isinstance(self.args[1], ExprMem) and isinstance(self.args[1].ptr, ExprOp):  # Major Opcodes #12
            # The second operand is an offset to a register
            o += " " + self.arg2str(self.args[0])
            o += ", %s" % ExprInt2SignedString(self.args[1].ptr.args[1])
            o += "(%s)" % self.arg2str(self.args[1].ptr.args[0])

        elif self.name == "SLL" and isinstance(self.args[1], ExprInt):  # Major Opcodes #6
            # The second operand is displayed in hex, not in decimal
            o += " " + self.arg2str(self.args[0])
            o += ", 0x%X" % int(self.args[1])

        elif self.name in ["ADD3", "SLT3"] and isinstance(self.args[2], ExprInt):
            o += " %s" % self.arg2str(self.args[0])
            o += ", %s" % self.arg2str(self.args[1])
            # The third operand is displayed in decimal, not in hex
            o += ", %s" % ExprInt2SignedString(self.args[2], pos_fmt="0x%X")

        elif self.name == "(RI)":
            return o

        else:
            args = []
            if self.args:
                o += " "
            for i, arg in enumerate(self.args):
                if not is_expr(arg):
                    raise ValueError('zarb arg type')
                x = self.arg2str(arg, pos=i)
                args.append(x)
            o += self.gen_args(args)

        return o

    def breakflow(self):
        """Instructions that stop a basic block."""

        if self.name in ["BRA", "BEQZ", "BNEZ", "BEQI", "BNEI", "BLTI", "BGEI", "BEQ", "BNE", "BSR"]:
            return True

        if self.name in ["JMP", "JSR", "RET"]:
            return True

        if self.name in ["RETI", "HALT", "SLEEP"]:
            return True

        return False

    def splitflow(self):
        """Instructions that splits a basic block, i.e. the CPU can go somewhere else."""

        if self.name in ["BEQZ", "BNEZ", "BEQI", "BNEI", "BLTI", "BGEI", "BEQ", "BNE", "BSR"]:
            return True

        return False

    def dstflow(self):
        """Instructions that explicitly provide the destination."""

        if self.name in ["BRA", "BEQZ", "BNEZ", "BEQI", "BNEI", "BLTI", "BGEI", "BEQ", "BNE", "BSR"]:
            return True

        if self.name in ["JMP"]:
            return True

        return False

    def dstflow2label(self, loc_db):
        """Set the label for the current destination.

           Note: it is used at disassembly"""

        if self.name == "JMP" and isinstance(self.args[0], ExprId):
            # 'JMP RM' does not provide the destination
            return

        # Compute the correct address
        num = self.get_dst_num()
        addr = int(self.args[num])
        if not self.name == "JMP":
            addr += self.offset

        # Get a new label at the address
        label = loc_db.get_or_create_offset_location(addr)

        # Assign the label to the correct instruction argument
        self.args[num] = ExprLoc(label, self.args[num].size)

    def get_dst_num(self):
        """Get the index of the argument that points to the instruction destination."""

        if self.name[-1] == "Z":
            num = 1
        elif self.name in ["BEQI", "BNEI", "BLTI", "BGEI", "BEQ", "BNE"]:
            num = 2
        else:
            num = 0

        return num

    def getdstflow(self, loc_db):
        """Get the argument that points to the instruction destination."""

        num = self.get_dst_num()
        return [self.args[num]]

    def is_subcall(self):
        """Instructions used to call sub functions."""

        return self.name in ["JSR", "BSR"]

    def fixDstOffset(self):
        """Fix/correct the instruction immediate according to the current offset

           Note: - it is used at assembly
                 - code inspired by miasm/arch/mips32/arch.py"""

        if self.name == "JMP" and isinstance(self.args[0], ExprInt):
            # 'JMP IMMEDIATE' does not need to be fixed
            return

        # Get the argument that needs to be fixed
        if not len(self.args):
            return
        num = self.get_dst_num()
        expr = self.args[num]

        # Check that the argument can be fixed
        if self.offset is None:
            raise ValueError("Symbol not resolved %s" % self.l)
        if not isinstance(expr, ExprInt):
            return

        # Adjust the immediate according to the current instruction offset
        off = expr.arg - self.offset
        if int(off % 2):
            raise ValueError("Strange offset! %r" % off)
        self.args[num] = ExprInt(off, 32)


class mep_additional_info(object):
    """Additional MeP instructions information
    """

    def __init__(self):
        self.except_on_instr = False


class mn_mep(cls_mn):
    """Toshiba MeP-c4 disassembler & assembler
    """

    # Define variables that stores information used to disassemble & assemble
    # Notes: - these variables are mandatory
    #        - they could be moved to the cls_mn class

    num = 0  # holds the number of mnemonics

    all_mn = list()  # list of mnenomnics, converted to metamn objects

    all_mn_mode = defaultdict(list)  # mneomnics, converted to metamn objects
                                     # Note:
                                     #   - the key is the mode # GV: what is it ?
                                     #   - the data is a list of mnemonics

    all_mn_name = defaultdict(list)  # mnenomnics strings
                                     # Note:
                                     #   - the key is the mnemonic string
                                     #   - the data is the corresponding
                                     #     metamn object

    all_mn_inst = defaultdict(list)  # mnemonics objects
                                     # Note:
                                     #   - the key is the mnemonic Python class
                                     #   - the data is an instantiated object

    bintree = dict()  # Variable storing internal values used to guess a
                      # mnemonic during disassembly

    # Defines the instruction set that will be used
    instruction = instruction_mep

    # Python module that stores registers information
    regs = mep_regs_module

    # Default delay slot
    # Note:
    #   - mandatory for the miasm Machine
    delayslot = 0

    # Architecture name
    name = "mep"

    # PC name depending on architecture attributes (here, l or b)
    pc = {'l': PC, 'b': PC}

    def additional_info(self):
        """Define instruction side effects # GV: not fully understood yet

        When used, it must return an object that implements specific
        variables, such as except_on_instr.

        Notes:
            - it must be implemented !
            - it could be moved to the cls_mn class
        """

        return mep_additional_info()

    @classmethod
    def gen_modes(cls, subcls, name, bases, dct, fields):
        """Ease populating internal variables used to disassemble & assemble, such
        as self.all_mn_mode, self.all_mn_name and self.all_mn_inst

        Notes:
            - it must be implemented !
            - it could be moved to the cls_mn class. All miasm architectures
              use the same code

        Args:
            cls: ?
            sublcs:
            name: mnemonic name
            bases: ?
            dct: ?
            fields: ?

        Returns:
            a list of ?

        """

        dct["mode"] = None
        return [(subcls, name, bases, dct, fields)]

    @classmethod
    def getmn(cls, name):
        """Get the mnemonic name

        Notes:
            - it must be implemented !
            - it could be moved to the cls_mn class. Most miasm architectures
              use the same code

        Args:
            cls:  the mnemonic class
            name: the mnemonic string
        """

        return name.upper()

    @classmethod
    def getpc(cls, attrib=None):
        """"Return the ExprId that represents the Program Counter.

        Notes:
            - mandatory for the symbolic execution
            - PC is defined in regs.py

        Args:
           attrib: architecture dependent attributes (here, l or b)
        """

        return PC

    @classmethod
    def getsp(cls, attrib=None):
        """"Return the ExprId that represents the Stack Pointer.

        Notes:
            - mandatory for the symbolic execution
            - SP is defined in regs.py

        Args:
           attrib: architecture dependent attributes (here, l or b)
        """

        return SP

    @classmethod
    def getbits(cls, bitstream, attrib, start, n):
        """Return an integer of n bits at the 'start' offset

           Note: code from miasm/arch/mips32/arch.py
        """

        # Return zero if zero bits are requested
        if not n:
            return 0

        o = 0  # the returned value
        while n:
            # Get a byte, the offset is adjusted according to the endianness
            offset = start // 8  # the offset in bytes
            n_offset = cls.endian_offset(attrib, offset)  # the adjusted offset
            c = cls.getbytes(bitstream, n_offset, 1)
            if not c:
                raise IOError

            # Extract the bits value
            c = ord(c)
            r = 8 - start % 8
            c &= (1 << r) - 1
            l = min(r, n)
            c >>= (r - l)
            o <<= l
            o |= c
            n -= l
            start += l

        return o

    @classmethod
    def endian_offset(cls, attrib, offset):
        """Adjust the byte offset according to the endianness"""

        if attrib == "l":  # Little Endian
            if offset % 2:
                return offset - 1
            else:
                return offset + 1

        elif attrib == "b":  # Big Endian
            return offset

        else:
            raise NotImplementedError("Bad MeP endianness")

    def value(self, mode):
        """Adjust the assembled instruction based on the endianness

           Note: code inspired by miasm/arch/mips32/arch.py
        """

        # Get the candidated
        candidates = super(mn_mep, self).value(mode)

        if mode == "l":
            # Invert bytes per 16-bits
            for i in range(len(candidates)):
                tmp = candidates[i][1] + candidates[i][0]
                if len(candidates[i]) == 4:
                    tmp += candidates[i][3] + candidates[i][2]
                candidates[i] = tmp
            return candidates

        elif mode == "b":
            return candidates

        else:
            raise NotImplementedError("Bad MeP endianness (%s)" % mode)


def addop(name, fields, args=None, alias=False):
    """Dynamically create the "name" object

    Notes:
        - it could be moved to a generic function such as:
          addop(name, fields, cls_mn, args=None, alias=False).
        - most architectures use the same code

    Args:
        name:   the mnemonic name
        fields: used to fill the object.__dict__'fields' attribute # GV: not understood yet
        args:   used to fill the object.__dict__'fields' attribute # GV: not understood yet
        alias:  used to fill the object.__dict__'fields' attribute # GV: not understood yet
    """

    namespace = {"fields": fields, "alias": alias}

    if args is not None:
        namespace["args"] = args

    # Dynamically create the "name" object
    type(name, (mn_mep,), namespace)


# Define specific operand parsers & converters

def deref2expr(s, l, parse_results):
    """Convert a parsed dereferenced register to an ExprMem"""

    # Only use the first results
    parse_results = parse_results[0]

    if type(parse_results[0]) == AstInt and isinstance(parse_results[2], AstId):
        return AstMem(parse_results[2] + parse_results[0], 32)  # 1 == "(" and 3 == ")"

    elif type(parse_results[0]) == int and isinstance(parse_results[2], AstId):
        return AstMem(parse_results[2] + AstOp('-', AstInt(-parse_results[0])), 32)  # 1 == "(" and 3 == ")"

    else:
        return AstMem(parse_results[1], 32)  # 0 == "(" and 2 == ")"


deref_reg_parser = Group(LPARENTHESIS + gpr_infos.parser + RPARENTHESIS).setParseAction(deref2expr)
deref_inc_reg_parser = Group(LPARENTHESIS + gpr_infos.parser + PLUSSIGN + RPARENTHESIS).setParseAction(deref2expr)
abs24_deref_parser = Group(LPARENTHESIS + HEX_INTEGER + RPARENTHESIS).setParseAction(deref2expr)
offset_deref_reg_parser = Group(HEX_INTEGER + LPARENTHESIS + gpr_infos.parser + RPARENTHESIS).setParseAction(deref2expr)

# Define registers decoders and encoders

class mep_arg(m_arg):
    def asm_ast_to_expr(self, arg, loc_db):
        """Convert AST to expressions

           Note: - code inspired by miasm/arch/mips32/arch.py"""

        if isinstance(arg, AstId):
            if isinstance(arg.name, ExprId):
                return arg.name
            if isinstance(arg.name, str) and arg.name in gpr_names:
                return None  # GV: why?
            loc_key = loc_db.get_or_create_name_location(arg.name)
            return ExprLoc(loc_key, 32)

        elif isinstance(arg, AstMem):
            addr = self.asm_ast_to_expr(arg.ptr, loc_db)
            if addr is None:
                return None
            return ExprMem(addr, 32)

        elif isinstance(arg, AstInt):
            return ExprInt(arg.value, 32)

        elif isinstance(arg, AstOp):
            args = [self.asm_ast_to_expr(tmp, loc_db) for tmp in arg.args]
            if None in args:
                return None
            return ExprOp(arg.op, *args)

        # Raise an exception if the argument was not processed
        message = "mep_arg.asm_ast_to_expr(): don't know what \
                   to do with a '%s' instance." % type(arg)
        raise Exception(message)

class mep_reg(reg_noarg, mep_arg):
    """Generic Toshiba MeP-c4 register

    Note:
        - the register size will be set using bs()
    """
    reg_info = gpr_infos  # the list of MeP-c4 registers defined in regs.py
    parser = reg_info.parser  # GV: not understood yet


class mep_deref_reg(mep_arg):
    """Generic Toshiba MeP-c4 dereferenced register

    Note:
        - the arg2str() method could be defined to change the output string
    """
    parser = deref_reg_parser

    def decode(self, v):
        """Transform the decoded value to a ExprMem(ExprId()) expression"""
        r = gpr_infos.expr[v]  # get the ExprId, i.e. the register expression
        self.expr = ExprMem(r, 32)
        return True

    def encode(self):
        """Ensure that we have a ExprMem(ExprId()) expression, and return the
        register value."""

        if not isinstance(self.expr, ExprMem):
            return False
        if not isinstance(self.expr.ptr, ExprId):
            return False

        # Get the ExprId index, i.e. its value
        self.value = gpr_exprs.index(self.expr.ptr)
        return True


class mep_reg_sp(mep_reg):
    """Dummy Toshiba MeP-c4 register that represents SP. It is used in
    instructions that implicitly use SP, such as ADD3.
    """
    implicit_reg = SP

    def decode(self, v):
        """Always return 'implicit_reg."""
        self.expr = self.implicit_reg
        return True

    def encode(self):
        """Do nothing"""
        return True


class mep_reg_tp(mep_reg_sp):
    """Dummy Toshiba MeP-c4 register that represents TP.
    """
    implicit_reg = TP


class mep_deref_reg_offset(mep_arg):
    """Toshiba MeP-c4 dereferenced register that represents SP, plus an
    offset.
    """
    parser = offset_deref_reg_parser

    def decode(self, v):
        """Modify the decoded value using the previously decoded
        register id.
        """

        # Apply the immediate mask
        se = sign_ext(v & 0xFFFF, 16, 32)  # GV: might not belong here
        int_id = ExprInt(se, 32)

        # Get the register expression
        reg_id = gpr_infos.expr[self.parent.reg04_deref.value]

        # Build the internal expression
        self.expr = ExprMem(reg_id + int_id, 32)

        return True

    def encode(self):
        """Modify the encoded value. One part is stored in this object, and
        the other one in reg04_deref.
        """

        # Verify the expression
        if not isinstance(self.expr, ExprMem):
            return False
        if not isinstance(self.expr.ptr, ExprOp):
            return False

        # Get the integer and check the upper bound
        v = int(self.expr.ptr.args[1]) & 0xFFFF

        # Encode the values
        self.parent.reg04_deref.value = gpr_exprs.index(self.expr.ptr.args[0])
        self.value = v & 0xFFFF
        return True


class mep_deref_sp_offset(mep_deref_reg):
    """Dummy Toshiba MeP-c4 dereferenced register that represents SP, plus an
    offset.
    Note: it is as generic as possible to ease its use in different instructions
    """
    implicit_reg = SP
    parser = offset_deref_reg_parser

    def decode(self, v):
        """Modify the decoded value using the previously decoded
        immediate.
        """

        immediate = None
        if getattr(self.parent, "imm7_align4", False):
            # Apply the immediate mask
            v = self.parent.imm7_align4.value & 0x1F

            # Shift value such as:
            #   imm7=iii_ii||00
            immediate = v << 2

        elif getattr(self.parent, "imm7", False):
            # Apply the immediate mask
            immediate = self.parent.imm7.value & 0x7F

        elif getattr(self.parent, "disp7_align2", False):
            # Apply the immediate mask
            disp7_align2 = self.parent.disp7_align2.value & 0x3F

            # Shift value such as:
            #   disp7 = ddd_ddd||0
            immediate = disp7_align2 << 1

        if immediate is not None:
            self.expr = ExprMem(self.implicit_reg + ExprInt(immediate, 32), 32)
            return True
        else:
            return False

    def encode(self):
        """Modify the encoded value. One part is stored in this object, and
        the other one in a parent immediate.
        """

        # Verify the expression
        if not isinstance(self.expr, ExprMem):
            return False
        if not isinstance(self.expr.ptr, ExprOp):
            return False
        if self.expr.ptr.args[0] != self.implicit_reg:
            return False

        if getattr(self.parent, "imm7_align4", False):

            # Get the integer and check the upper bound
            v = int(self.expr.ptr.args[1].arg)
            if v > 0x80:
                return False

            # Encode the value
            self.parent.imm7_align4.value = v >> 2

            return True

        elif getattr(self.parent, "imm7", False):

            # Get the integer and check the upper bound
            v = int(self.expr.ptr.args[1].arg)
            if v > 0x80:
                return False

            # Encode the value
            self.parent.imm7.value = v

            return True

        elif getattr(self.parent, "disp7_align2", False):

            # Get the integer and check the upper bound
            v = int(self.expr.ptr.args[1].arg)
            if v > 0x80:
                return False

            # Encode the value
            self.parent.disp7_align2.value = v >> 1

            return True

        return False


class mep_deref_tp_offset(mep_deref_sp_offset):
    """Dummy Toshiba MeP-c4 dereferenced register that represents TP, plus an
    offset.
    """
    implicit_reg = TP


class mep_copro_reg(reg_noarg, mep_arg):
    """Generic Toshiba MeP-c4 coprocessor register
    """
    reg_info = copro_gpr_infos  # the list of MeP-c4 coprocessor registers defined in regs.py
    parser = reg_info.parser  # GV: not understood yet


class mep_copro_reg_split(mep_copro_reg):
    """Generic Toshiba MeP-c4 coprocessor register encode into different fields
    """

    def decode(self, v):
        """Modify the decoded value using the previously decoded imm4_noarg.
        """

        # Apply the immediate mask
        v = v & self.lmask

        # Shift values such as:
        #   CRn=NNnnnn
        crn = (v << 4) + (self.parent.imm4.value & 0xF)

        # Build the internal expression
        self.expr = ExprId("C%d" % crn, 32)
        return True

    def encode(self):
        """Modify the encoded value. One part is stored in this object, and
        the other one in imm4_noarg.
        """

        if not isinstance(self.expr, ExprId):
            return False

        # Get the register and check the upper bound
        reg_name = self.expr.name
        if reg_name[0] != "C":
            return False
        reg_value = copro_gpr_names.index(reg_name)
        if reg_value > 0x3f:
            return False

        # Encode the value into two parts
        self.parent.imm4.value = (reg_value & 0xF)
        self.value = (reg_value >> 4) & 0x3
        return True


class mep_deref_inc_reg(mep_deref_reg):
    """Generic Toshiba MeP-c4 coprocess dereferenced & incremented register
    """
    parser = deref_inc_reg_parser


# Immediate decoders and encoders

class mep_int32_noarg(int32_noarg):
    """Generic Toshiba MeP-c4 signed immediate

       Note: encode() is copied from int32_noarg.encode() and modified to allow
             small (< 32 bits) signed immediate to be manipulated.

    """

    def encode(self):
        if not isinstance(self.expr, ExprInt):
            return False
        v = int(self.expr)
        # Note: the following lines were commented on purpose
        #if sign_ext(v & self.lmask, self.l, self.intsize) != v:
        #    return False
        v = self.encodeval(v & self.lmask)
        self.value = v & self.lmask
        return True


class mep_imm(imm_noarg, mep_arg):
    """Generic Toshiba MeP-c4 immediate

    Note:
        - the immediate size will be set using bs()
    """
    parser = base_expr


class mep_imm6(mep_int32_noarg):
    """Toshiba MeP-c4 signed 6 bits immediate."""
    parser = base_expr
    intsize = 6
    intmask = (1 << intsize) - 1
    int2expr = lambda self, x: ExprInt(sign_ext(x, self.l, 32), 32)


class mep_imm8(mep_int32_noarg):
    """Toshiba MeP-c4 signed 8 bits immediate."""
    parser = base_expr
    intsize = 8
    intmask = (1 << intsize) - 1
    int2expr = lambda self, x: ExprInt(sign_ext(x, self.l, 32), 32)


class mep_imm16(mep_int32_noarg):
    """Toshiba MeP-c4 16 bits immediate."""
    parser = base_expr
    intsize = 16
    intmask = (1 << intsize) - 1
    int2expr = lambda self, x: ExprInt(x, 32)


class mep_imm16_signed(mep_int32_noarg):
    """Toshiba MeP-c4 signed 16 bits immediate."""
    parser = base_expr
    intsize = 16
    intmask = (1 << intsize) - 1
    int2expr = lambda self, x: ExprInt(sign_ext(x, self.l, 32), 32)


class mep_target24(mep_imm):
    """Toshiba MeP-c4 target24 immediate, as used in JMP
    """

    def decode(self, v):
        """Modify the decoded value using the previously decoded imm7.
        """

        # Apply the immediate mask
        v = v & self.lmask

        # Shift values such as:
        #   target24=tttt_tttt_tttt_tttt||TTT_TTTT||0
        target24 = (v << 8) + ((self.parent.imm7.value & 0x7F) << 1)

        # Build the internal expression
        self.expr = ExprInt(target24, 32)
        return True

    def encode(self):
        """Modify the encoded value. One part is stored in this object, and
        the other one in imm7.
        """

        if not isinstance(self.expr, ExprInt):
            return False

        # Get the integer and apply a mask
        v = int(self.expr) & 0x00FFFFFF

        # Encode the value into two parts
        self.parent.imm7.value = (v & 0xFF) >> 1
        self.value = v >> 8
        return True


class mep_target24_signed(mep_target24):
    """Toshiba MeP-c4 target24 signed immediate, as used in BSR
    """

    def decode(self, v):
        """Perform sign extension
        """

        mep_target24.decode(self, v)
        v = int(self.expr)
        self.expr = ExprInt(sign_ext(v, 24, 32), 32)

        return True


class mep_code20(mep_imm):
    """Toshiba MeP-c4 code20 immediate, as used in DSP1
    """

    def decode(self, v):
        """Modify the decoded value using the previously decoded imm4_noarg.
        """

        # Apply the immediate mask
        v = v & self.lmask

        # Shift values such as:
        #   code20=mmmm_cccc_cccc_cccc_cccc
        code20 = v + ((self.parent.imm4.value & 0xFF) << 16)

        # Build the internal expression
        self.expr = ExprInt(code20, 32)
        return True

    def encode(self):
        """Modify the encoded value. One part is stored in this object, and
        the other one in imm4_noarg.
        """

        if not isinstance(self.expr, ExprInt):
            return False

        # Get the integer and check the upper bound
        v = int(self.expr.arg)
        if v > 0xffffff:
            return False

        # Encode the value into two parts
        self.parent.imm4 = ((v >> 16) & 0xFF)
        self.value = v
        return True


class mep_code24(mep_imm):
    """Toshiba MeP-c4 code24 immediate, as used in CP
    """

    def decode(self, v):
        """Modify the decoded value using the previously decoded imm8_CCCC_CCCC.
        """

        # Shift values such as:
        #   code24=CCCC_CCCC||cccc_cccc_cccc_cccc
        code24 = v + ((self.parent.imm8_CCCC_CCCC.value & 0xFF) << 16)

        # Build the internal expression
        self.expr = ExprInt(code24, 32)
        return True

    def encode(self):
        """Modify the encoded value. One part is stored in this object, and
        the other one in imm8_CCCC_CCCC.
        """

        if not isinstance(self.expr, ExprInt):
            return False

        # Get the integer and check the upper bound
        v = int(self.expr.arg)
        if v > 0xFFFFFF:
            return False

        # Encode the value into two parts
        self.parent.imm8_CCCC_CCCC.value = ((v >> 16) & 0xFF)
        self.value = v & 0xFFFF
        return True


class mep_imm7_align4(mep_imm):
    """Toshiba MeP-c4 imm7.align4 immediate, as used in Major #4 opcodes
    """

    def decode(self, v):
        """Modify the decoded value.
        """

        # Apply the immediate mask
        v = v & self.lmask

        # Shift value such as:
        #   imm7=iii_ii||00
        imm7_align4 = v << 2

        # Build the internal expression
        self.expr = ExprInt(imm7_align4, 32)
        return True

    def encode(self):
        """Modify the encoded value.
        """

        if not isinstance(self.expr, ExprInt):
            return False

        # Get the integer and check the upper bound
        v = int(self.expr)
        if v > 0x80:
            return False

        # Encode the value
        self.value = v >> 2
        return True


class mep_imm5_Iiiii (mep_imm):
    """Toshiba MeP-c4 imm5 immediate, as used in STC & LDC. It encodes a
    control/special register.
    """

    reg_info = csr_infos  # the list of MeP-c4 control/special registers defined in regs.py
    parser = reg_info.parser  # GV: not understood yet

    def decode(self, v):
        """Modify the decoded value using the previously decoded imm4_iiii
        """

        # Apply the immediate mask
        I = v & self.lmask

        # Shift values such as:
        #   imm5=I||iiii
        imm5 = (I << 4) + (self.parent.imm4_iiii.value & 0xF)

        # Build the internal register expression
        self.expr = ExprId(csr_names[imm5], 32)
        return True

    def encode(self):
        """Modify the encoded value. One part is stored in this object, and
        the other one in imm4_iiii.
        """

        if not isinstance(self.expr, ExprId):
            return False

        # Get the register number and check the upper bound
        v = csr_names.index(self.expr.name)
        if v > 0x1F:
            return False

        # Encode the value into two parts
        self.parent.imm4_iiii.value = v & 0xF  # iiii
        self.value = (v >> 4) & 0b1  # I
        return True


class mep_disp7_align2(mep_imm):
    """Toshiba MeP-c4 disp7.align2 immediate, as used in Major #8 opcodes
    """
    upper_bound = 0x7F
    bits_shift = 1

    def decode(self, v):
        """Modify the decoded value.
        """

        # Apply the immediate mask
        v = v & self.lmask

        # Shift value such as:
        #   disp7 = ddd_ddd||0
        disp7_align2 = (v << self.bits_shift)

        # Sign extension
        disp7_align2 = sign_ext(disp7_align2, self.l + self.bits_shift, 32)

        # Build the internal expression
        self.expr = ExprInt(disp7_align2, 32)
        return True

    def encode(self):
        """Modify the encoded value.
        """

        if not isinstance(self.expr, ExprInt):
            return False

        # Get the integer
        v = int(self.expr) & self.upper_bound

        # Encode the value
        self.value = (v >> self.bits_shift) & self.upper_bound
        self.value = (v & self.upper_bound) >> self.bits_shift
        return True


class mep_disp8_align2(mep_disp7_align2):
    upper_bound = 0xFF


class mep_disp8_align4(mep_disp7_align2):
    upper_bound = 0xFF
    bits_shift = 2


class mep_imm8_align8(mep_disp7_align2):
    upper_bound = 0xFF
    bits_shift = 3


class mep_disp12_align2(mep_disp7_align2):
    upper_bound = 0xFFF


class mep_disp12_align2_signed(mep_disp12_align2):

    def decode(self, v):
        """Perform sign extension.
        """
        mep_disp12_align2.decode(self, v)
        v = int(self.expr)

        self.expr = ExprInt(sign_ext(v, 12, 32), 32)
        return True


class mep_disp17(mep_disp7_align2):
    upper_bound = 0x1FFFF


class mep_imm24(mep_imm):
    """Toshiba MeP-c4 imm24 immediate, as used in MOVU
    """

    def decode(self, v):
        """Modify the decoded value.
        """

        # Apply the immediate mask
        v = v & self.lmask

        # Shift values such as:
        #   imm24=iiii_iiii_iiii_iiii||IIII_IIIII
        imm24 = ((v & 0xFFFF) << 8) + ((v & 0xFF0000) >> 16)

        # Build the internal expression
        self.expr = ExprInt(imm24, 32)
        return True

    def encode(self):
        """Modify the encoded value.
        """

        if not isinstance(self.expr, ExprInt):
            return False

        # Get the integer and check the upper bound
        v = int(self.expr)
        if v > 0xFFFFFF:
            return False

        # Encode the value
        self.value = ((v & 0xFFFF00) >> 8) + ((v & 0xFF) << 16)
        return True


class mep_abs24(mep_imm):
    """Toshiba MeP-c4 abs24 immediate
    """
    parser = abs24_deref_parser

    def decode(self, v):
        """Modify the decoded value using the previously decoded imm6.
        """

        # Apply the immediate mask
        v = v & self.lmask

        # Shift values such as:
        #   abs24=dddd_dddd_dddd_dddd||DDDD_DD||00
        abs24 = (v << 8) + ((self.parent.imm6.value & 0x3F) << 2)

        # Build the internal expression
        self.expr = ExprMem(ExprInt(abs24, 32), 32)
        return True

    def encode(self):
        """Modify the encoded value. One part is stored in this object, and
        the other one in imm6.
        """

        if not (isinstance(self.expr, ExprMem) and isinstance(self.expr.ptr, ExprInt)):
            return False

        # Get the integer and check the upper bound
        v = int(self.expr.ptr)
        if v > 0xffffff:
            return False

        # Encode the value into two parts
        self.parent.imm6.value = (v & 0xFF) >> 2
        self.value = v >> 8
        return True


# Define MeP-c4 assembly operands

reg04 = bs(l=4,  # length in bits
           cls=(mep_reg, ))  # class implementing decoding & encoding

reg04_l = bs(l=4, cls=(mep_reg, ))

reg04_m = bs(l=4, cls=(mep_reg, ))

reg04_n = bs(l=4, cls=(mep_reg, ))

reg00 = bs(l=0, cls=(mep_reg, ))

reg00_sp = bs(l=0, cls=(mep_reg_sp, ))

reg00_tp = bs(l=0, cls=(mep_reg_tp, ))

reg00_deref_sp = bs(l=0, cls=(mep_deref_sp_offset, ))

reg00_deref_tp = bs(l=0, cls=(mep_deref_tp_offset, ))

reg03 = bs(l=3, cls=(mep_reg, ))

reg04_deref = bs(l=4, cls=(mep_deref_reg,))

reg04_deref_noarg = bs(l=4, fname="reg04_deref")

reg04_inc_deref = bs(l=4, cls=(mep_deref_inc_reg,))

copro_reg04 = bs(l=4, cls=(mep_copro_reg,))

copro_reg05 = bs(l=1, cls=(mep_copro_reg_split,))

copro_reg06 = bs(l=2, cls=(mep_copro_reg_split,))

disp2 = bs(l=2, cls=(mep_imm, ))

imm2 = disp2

imm3 = bs(l=3, cls=(mep_imm, ))

imm4 = bs(l=4, cls=(mep_imm, ))

imm4_noarg = bs(l=4, fname="imm4")

imm4_iiii_noarg = bs(l=4, fname="imm4_iiii")

imm5 = bs(l=5, cls=(mep_imm, ))

imm5_Iiiii = bs(l=1, cls=(mep_imm5_Iiiii, ))  # it is not an immediate, but a
                                              # control/special register.

imm6 = bs(l=6, cls=(mep_imm6, mep_arg))

imm6_noarg = bs(l=6, fname="imm6")

imm7 = bs(l=7, cls=(mep_imm, ))

imm7_noarg = bs(l=7, fname="imm7")  # Note:
                                    #   - will be decoded as a 7 bits immediate
                                    #   - fname is used to set the operand name
                                    #     used in mep_target24 to merge operands
                                    #     values. By default, the bs class fills
                                    #     fname with an hex string compute from
                                    #     arguments passed to __init__

imm7_align4 = bs(l=5, cls=(mep_imm7_align4,))

imm7_align4_noarg = bs(l=5, fname="imm7_align4")

disp7_align2 = bs(l=6, cls=(mep_disp7_align2,))

disp7_align2_noarg = bs(l=6, fname="disp7_align2")

imm8 = bs(l=8, cls=(mep_imm8, mep_arg))

imm8_noarg = bs(l=8, fname="imm8_CCCC_CCCC")

disp8 = bs(l=7, cls=(mep_disp8_align2, ))

imm8_align2 = bs(l=7, cls=(mep_disp8_align2, ))

imm8_align4 = bs(l=6, cls=(mep_disp8_align4, ))

imm8_align8 = bs(l=5, cls=(mep_imm8_align8, ))

imm12 = bs(l=12, cls=(mep_imm, ))

disp12_signed = bs(l=11, cls=(mep_disp12_align2_signed, ))

imm16 = bs(l=16, cls=(mep_imm16, mep_arg))
imm16_signed = bs(l=16, cls=(mep_imm16_signed, mep_arg))

disp16_reg_deref = bs(l=16, cls=(mep_deref_reg_offset,))

disp17 = bs(l=16, cls=(mep_disp17, ))

imm18 = bs(l=19, cls=(mep_imm, ))

imm_code20 = bs(l=16, cls=(mep_code20, ))

imm24 = bs(l=24, cls=(mep_imm24, ))

imm_target24 = bs(l=16, cls=(mep_target24, ))
imm_target24_signed = bs(l=16, cls=(mep_target24_signed, ))

imm_code24 = bs(l=16, cls=(mep_code24, ))

abs24 = bs(l=16, cls=(mep_abs24, ))


# MeP-c4 mnemonics objects

### <Major Opcode #0>

# MOV Rn,Rm - 0000_nnnn_mmmm_0000
addop("MOV", [bs("0000"), reg04, reg04, bs("0000")])

# NEG Rn,Rm - 0000_nnnn_mmmm_0001
addop("NEG", [bs("0000"), reg04, reg04, bs("0001")])

# SLT3 R0,Rn,Rm - 0000_nnnn_mmmm_0010
addop("SLT3", [bs("0000"), reg00, reg04, reg04, bs("0010")])

# SLTU3 R0,Rn,Rm - 0000_nnnn_mmmm_0011
addop("SLTU3", [bs("0000"), reg00, reg04, reg04, bs("0011")])

# SUB Rn,Rm - 0000_nnnn_mmmm_0100
addop("SUB", [bs("0000"), reg04, reg04, bs("0100")])

# SBVCK3 R0,Rn,Rm - 0000_nnnn_mmmm_0101
addop("SBVCK3", [bs("0000"), reg00, reg04, reg04, bs("0101")])

# (RI) - 0000_xxxx_xxxx_0110
addop("(RI)", [bs("0000"), reg04, reg04, bs("0110")])

# ADVCK3 R0,Rn,Rm - 0000_nnnn_mmmm_0111
addop("ADVCK3", [bs("0000"), reg00, reg04, reg04, bs("0111")])

# SB Rn,(Rm) - 0000_nnnn_mmmm_1000
addop("SB", [bs("0000"), reg04, reg04_deref, bs("1000")])

# SH Rn,(Rm) - 0000_nnnn_mmmm_1001
addop("SH", [bs("0000"), reg04, reg04_deref, bs("1001")])

# SW Rn,(Rm) - 0000_nnnn_mmmm_1010
addop("SW", [bs("0000"), reg04, reg04_deref, bs("1010")])

# LBU Rn,(Rm) - 0000_nnnn_mmmm_1011
addop("LBU", [bs("0000"), reg04, reg04_deref, bs("1011")])

# LB Rn,(Rm) - 0000_nnnn_mmmm_1100
addop("LB", [bs("0000"), reg04, reg04_deref, bs("1100")])

# LH Rn,(Rm) - 0000_nnnn_mmmm_1101
addop("LH", [bs("0000"), reg04, reg04_deref, bs("1101")])

# LW Rn,(Rm) - 0000_nnnn_mmmm_1110
addop("LW", [bs("0000"), reg04, reg04_deref, bs("1110")])

# LHU Rn,(Rm) - 0000_nnnn_mmmm_1111
addop("LHU", [bs("0000"), reg04, reg04_deref, bs("1111")])


### <Major Opcode #1>

# OR Rn,Rm - 0001_nnnn_mmmm_0000
addop("OR", [bs("0001"), reg04, reg04, bs("0000")])

# AND Rn,Rm - 0001_nnnn_mmmm_0001
addop("AND", [bs("0001"), reg04, reg04, bs("0001")])

# XOR Rn,Rm - 0001_nnnn_mmmm_0010
addop("XOR", [bs("0001"), reg04, reg04, bs("0010")])

# NOR Rn,Rm - 0001_nnnn_mmmm_0011
addop("NOR", [bs("0001"), reg04, reg04, bs("0011")])

# MUL Rn,Rm - 0001_nnnn_mmmm_0100
addop("MUL", [bs("0001"), reg04, reg04, bs("0100")])

# MULU Rn,Rm - 0001_nnnn_mmmm_0101
addop("MULU", [bs("0001"), reg04, reg04, bs("0101")])

# MULR Rn,Rm - 0001_nnnn_mmmm_0110
addop("MULR", [bs("0001"), reg04, reg04, bs("0110")])

# MULRU Rn,Rm - 0001_nnnn_mmmm_0111
addop("MULRU", [bs("0001"), reg04, reg04, bs("0111")])

# DIV Rn,Rm - 0001_nnnn_mmmm_1000
addop("DIV", [bs("0001"), reg04, reg04, bs("1000")])

# DIVU Rn,Rm - 0001_nnnn_mmmm_1001
addop("DIVU", [bs("0001"), reg04, reg04, bs("1001")])

# (RI) - 0001_xxxx_xxxx_1010
addop("(RI)", [bs("0001"), reg04, reg04, bs("1010")])

# (RI) - 0001_xxxx_xxxx_1011
addop("(RI)", [bs("0001"), reg04, reg04, bs("1011")])

# SSARB disp2(Rm) - 0001_00dd_mmmm_1100
addop("SSARB", [bs("000100"), disp2, reg04_deref, bs("1100")])

# EXTB Rn - 0001_nnnn_0000_1101
addop("EXTB", [bs("0001"), reg04, bs("00001101")])

# EXTH Rn - 0001_nnnn_0010_1101
addop("EXTH", [bs("0001"), reg04, bs("00101101")])

# EXTUB Rn - 0001_nnnn_1000_1101
addop("EXTUB", [bs("0001"), reg04, bs("10001101")])

# EXTUH Rn - 0001_nnnn_1010_1101
addop("EXTUH", [bs("0001"), reg04, bs("10101101")])

# JMP Rm - 0001_0000_mmmm_1110
addop("JMP", [bs("00010000"), reg04, bs("1110")])

# JSR Rm - 0001_0000_mmmm_1111
addop("JSR", [bs("00010000"), reg04, bs("1111")])

# JSRV Rm - 0001_1000_mmmm_1111
addop("JSRV", [bs("00011000"), reg04, bs("1111")])


### <Major Opcode #2>

# BSETM (Rm),imm3 - 0010_0iii_mmmm_0000
addop("BSETM", [bs("00100"), imm3, reg04_deref, bs("0000")], [reg04_deref, imm3])

# BCLRM (Rn),imm3 - 0010_0iii_mmmm_0001
addop("BCLRM", [bs("00100"), imm3, reg04_deref, bs("0001")], [reg04_deref, imm3])

# BNOTM (Rm),imm3 - 0010_0iii_mmmm_0010
addop("BNOTM", [bs("00100"), imm3, reg04_deref, bs("0010")], [reg04_deref, imm3])

# BTSTM R0,(Rm),imm3 - 0010_0iii_mmmm_0011
addop("BTSTM", [bs("00100"), reg00, imm3, reg04_deref, bs("0011")], [reg00, reg04_deref, imm3])

# TAS Rn,(Rm) - 0010_nnnn_mmmm_0100
addop("TAS", [bs("0010"), reg04, reg04_deref, bs("0100")])

# (RI) - 0010_xxxx_xxxx_0101
addop("(RI)", [bs("0010"), reg04, reg04, bs("0101")])

# SL1AD3 R0,Rn,Rm - 0010_nnnn_mmmm_0110
addop("SL1AD3", [bs("0010"), reg00, reg04, reg04, bs("0110")])

# SL2AD3 R0,Rn,Rm - 0010_nnnn_mmmm_0111
addop("SL2AD3", [bs("0010"), reg00, reg04, reg04, bs("0111")])

# (RI) - 0010_xxxx_xxxx_1000
addop("(RI)", [bs("0010"), reg04, reg04, bs("1000")])

# (RI) - 0010_xxxx_xxxx_1001
addop("(RI)", [bs("0010"), reg04, reg04, bs("1001")])

# (RI) - 0010_xxxx_xxxx_1010
addop("(RI)", [bs("0010"), reg04, reg04, bs("1010")])

# (RI) - 0010_xxxx_xxxx_1011
addop("(RI)", [bs("0010"), reg04, reg04, bs("1011")])

# SRL Rn,Rm - 0010_nnnn_mmmm_1100
addop("SRL", [bs("0010"), reg04, reg04, bs("1100")])

# SRA Rn,Rm - 0010_nnnn_mmmm_1101
addop("SRA", [bs("0010"), reg04, reg04, bs("1101")])

# SLL Rn,Rm - 0010_nnnn_mmmm_1110
addop("SLL", [bs("0010"), reg04, reg04, bs("1110")])

# FSFT Rn,Rm - 0010_nnnn_mmmm_1111
addop("FSFT", [bs("0010"), reg04, reg04, bs("1111")])


### <Major Opcode #3>

# SWCPI CRn,(Rm+) - 0011_nnnn_mmmm_0000
addop("SWCPI", [bs("0011"), copro_reg04, reg04_inc_deref, bs("0000")])

# LWCPI CRn,(Rm+) - 0011_nnnn_mmmm_0001
addop("LWCPI", [bs("0011"), copro_reg04, reg04_inc_deref, bs("0001")])

# SMCPI CRn,(Rm+) - 0011_nnnn_mmmm_0010
addop("SMCPI", [bs("0011"), copro_reg04, reg04_inc_deref, bs("0010")])

# LMCPI CRn,(Rm+) - 0011_nnnn_mmmm_0011
addop("LMCPI", [bs("0011"), copro_reg04, reg04_inc_deref, bs("0011")])

# SWCP CRn,(Rm) - 0011_nnnn_mmmm_1000
addop("SWCP", [bs("0011"), copro_reg04, reg04_deref, bs("1000")])

# LWCP CRn,(Rm) - 0011_nnnn_mmmm_1001
addop("LWCP", [bs("0011"), copro_reg04, reg04_deref, bs("1001")])

# SMCP CRn,(Rm) - 0011_nnnn_mmmm_1010
addop("SMCP", [bs("0011"), copro_reg04, reg04_deref, bs("1010")])

# LMCP CRn,(Rm) - 0011_nnnn_mmmm_1011
addop("LMCP", [bs("0011"), copro_reg04, reg04_deref, bs("1011")])


### <Major Opcode #4>

# ADD3 Rn,SP,imm7.align4 - 0100_nnnn_0iii_ii00
addop("ADD3", [bs("0100"), reg04, reg00_sp, bs("0"), imm7_align4, bs("00")])

# SW Rn,disp7.align4(SP) - 0100_nnnn_0ddd_dd10
# Note: disp7.align4 is the same as imm7.align4
addop("SW", [bs("0100"), reg04, bs("0"), imm7_align4_noarg, reg00_deref_sp, bs("10")])

# LW Rn,disp7.align4(SP) - 0100_nnnn_0ddd_dd11
addop("LW", [bs("0100"), reg04, bs("0"), imm7_align4_noarg, reg00_deref_sp, bs("11")])

# SW Rn[0-7],disp7.align4(TP) - 0100_0nnn_1ddd_dd10
addop("SW", [bs("01000"), reg03, bs("1"), imm7_align4_noarg, reg00_deref_tp, bs("10")])

# LW Rn[0-7],disp7.align4(TP) - 0100_0nnn_1ddd_dd11
addop("LW", [bs("01000"), reg03, bs("1"), imm7_align4_noarg, reg00_deref_tp, bs("11")])

# LBU Rn[0-7],disp7(TP) - 0100_1nnn_1ddd_dddd
addop("LBU", [bs("01001"), reg03, bs("1"), imm7_noarg, reg00_deref_tp], [reg03, reg00_deref_tp])

### <Major Opcode #5>

# MOV Rn,imm8 - 0101_nnnn_iiii_iiii
addop("MOV", [bs("0101"), reg04, imm8])


### <Major Opcode #6>

# ADD Rn,imm6 - 0110_nnnn_iiii_ii00
addop("ADD",  # mnemonic name
      [bs("0110"), reg04, imm6, bs("00")])  # mnemonic description

# SLT3 R0,Rn,imm5 - 0110_nnnn_iiii_i001
addop("SLT3", [bs("0110"), reg00, reg04, imm5, bs("001")])

# SRL Rn,imm5 - 0110_nnnn_iiii_i010
addop("SRL", [bs("0110"), reg04, imm5, bs("010")])

# SRA Rn,imm5 - 0110_nnnn_iiii_i011
addop("SRA", [bs("0110"), reg04, imm5, bs("011")])

# SLTU3 R0,Rn,imm5 - 0110_nnnn_iiii_i101
addop("SLTU3", [bs("0110"), reg00, reg04, imm5, bs("101")])

# SLL Rn,imm5 - 0110_nnnn_iiii_i110
addop("SLL", [bs("0110"), reg04, imm5, bs("110")])

# SLL3 R0,Rn,imm5 - 0110_nnnn_iiii_i111
addop("SLL3", [bs("0110"), reg00, reg04, imm5, bs("111")])


### <Major Opcode #7>

# DI - 0111_0000_0000_0000
addop("DI", [bs("0111000000000000")])

# EI - 0111_0000_0001_0000
addop("EI", [bs("0111000000010000")])

# SYNCM - 0111_0000_0001_0001
addop("SYNCM", [bs("0111000000010001")])

# SYNCCP - 0111_0000_0010_0001
addop("SYNCCP", [bs("0111000000100001")])

# RET - 0111_0000_0000_0010
addop("RET", [bs("0111000000000010")])

# RETI - 0111_0000_0001_0010
addop("RETI", [bs("0111000000010010")])

# HALT - 0111_0000_0010_0010
addop("HALT", [bs("0111000000100010")])

# BREAK - 0111_0000_0011_0010
addop("BREAK", [bs("0111000000110010")])

# SLEEP - 0111_0000_0110_0010
addop("SLEEP", [bs("0111000001100010")])

# DRET - 0111_0000_0001_0011
addop("DRET", [bs("0111000000010011")])

# DBREAK - 0111_0000_0011_0011
addop("DBREAK", [bs("0111000000110011")])

# CACHE imm4,(Rm) - 0111_iiii_mmmm_0100
addop("CACHE", [bs("0111"), imm4, reg04_deref, bs("0100")])

# (RI) - 0111_xxxx_xxxx_0101
addop("(RI)", [bs("0111"), reg04, reg04, bs("0101")])

# SWI imm2 - 0111_0000_00ii_0110
addop("SWI", [bs("0111000000"), imm2, bs("0110")])

# (RI) - 0111_xxxx_xxxx_0111
addop("(RI)", [bs("0111"), reg04, reg04, bs("0111")])

# STC Rn,imm5 - 0111_nnnn_iiii_100I
addop("STC", [bs("0111"), reg04, imm4_iiii_noarg, bs("100"), imm5_Iiiii])

# LDC Rn,imm5 - 0111_nnnn_iiii_101I
addop("LDC", [bs("0111"), reg04, imm4_iiii_noarg, bs("101"), imm5_Iiiii])

# (RI) - 0111_xxxx_xxxx_1100
addop("(RI)", [bs("0111"), reg04, reg04, bs("1100")])

# (RI) - 0111_xxxx_xxxx_1101
addop("(RI)", [bs("0111"), reg04, reg04, bs("1101")])

# (RI) - 0111_xxxx_xxxx_1110
addop("(RI)", [bs("0111"), reg04, reg04, bs("1110")])

# (RI) - 0111_xxxx_xxxx_1111
addop("(RI)", [bs("0111"), reg04, reg04, bs("1111")])


### <Major Opcode #8>

# SB Rn[0-7],disp7(TP) - 1000_0nnn_0ddd_dddd
addop("SB", [bs("10000"), reg03, bs("0"), imm7_noarg, reg00_deref_tp])

# SH Rn[0-7],disp7.align2(TP) - 1000_0nnn_1ddd_ddd0
# (disp7.align2 = ddd_ddd||0)
addop("SH", [bs("10000"), reg03, bs("1"), disp7_align2_noarg, bs("0"), reg00_deref_tp])

# LB Rn[0-7],disp7(TP) - 1000_1nnn_0ddd_dddd
addop("LB", [bs("10001"), reg03, bs("0"), imm7_noarg, reg00_deref_tp])

# LH Rn[0-7],disp7.align2(TP) - 1000_1nnn_1ddd_ddd0
addop("LH", [bs("10001"), reg03, bs("1"), disp7_align2_noarg, bs("0"), reg00_deref_tp])

# LHU Rn[0-7],disp7.align2(TP) - 1000_1nnn_1ddd_ddd1
addop("LHU", [bs("10001"), reg03, bs("1"), disp7_align2_noarg, bs("1"), reg00_deref_tp])


### <Major Opcode #9>

# ADD3 Rl,Rn,Rm - 1001_nnnn_mmmm_llll
addop("ADD3", [bs("1001"), reg04_n, reg04_m, reg04_l], [reg04_l, reg04_n, reg04_m])


### <Major Opcode #10>

# BEQZ Rn,disp8.align2 - 1010_nnnn_dddd_ddd0
# (disp8=dddd_ddd||0)
addop("BEQZ", [bs("1010"), reg04, disp8, bs("0")])

# BNEZ Rn,disp8.align2 - 1010_nnnn_dddd_ddd1
addop("BNEZ", [bs("1010"), reg04, disp8, bs("1")])


### <Major Opcode #11>

# BRA disp12.align2 - 1011_dddd_dddd_ddd0
# (disp12=dddd_dddd_ddd||0)
addop("BRA", [bs("1011"), disp12_signed, bs("0")])

# BSR disp12.align2 - 1011_dddd_dddd_ddd1
addop("BSR", [bs("1011"), disp12_signed, bs("1")])


### <Major Opcode #12>

# ADD3 Rn,Rm,imm16 - 1100_nnnn_mmmm_0000 iiii_iiii_iiii_iiii
addop("ADD3", [bs("1100"), reg04, reg04, bs("0000"), imm16_signed])

# MOV Rn,imm16 - 1100_nnnn_0000_0001 iiii_iiii_iiii_iiii
addop("MOV", [bs("1100"), reg04, bs("00000001"), imm16])

# MOVU Rn,imm16 - 1100_nnnn_0001_0001 iiii_iiii_iiii_iiii
addop("MOVU", [bs("1100"), reg04, bs("00010001"), imm16])

# MOVH Rn,imm16 - 1100_nnnn_0010_0001 iiii_iiii_iiii_iiii
addop("MOVH", [bs("1100"), reg04, bs("00100001"), imm16])

# SLT3 Rn,Rm,imm16 - 1100_nnnn_mmmm_0010 iiii_iiii_iiii_iiii
addop("SLT3", [bs("1100"), reg04, reg04, bs("0010"), imm16_signed])

# SLTU3 Rn,Rm,imm16 - 1100_nnnn_mmmm_0011 iiii_iiii_iiii_iiii
addop("SLTU3", [bs("1100"), reg04, reg04, bs("0011"), imm16])

# OR3 Rn,Rm,imm16 - 1100_nnnn_mmmm_0100 iiii_iiii_iiii_iiii
addop("OR3", [bs("1100"), reg04, reg04, bs("0100"), imm16])

# AND3 Rn,Rm,imm16 - 1100_nnnn_mmmm_0101 iiii_iiii_iiii_iiii
addop("AND3", [bs("1100"), reg04, reg04, bs("0101"), imm16])

# XOR3 Rn,Rm,imm16 - 1100_nnnn_mmmm_0110 iiii_iiii_iiii_iiii
addop("XOR3", [bs("1100"), reg04, reg04, bs("0110"), imm16])

# (RI) - 1100_xxxx_xxxx_0111 xxxx_xxxx_xxxx_xxxx
addop("(RI)", [bs("1100"), imm8, bs("0111"), imm16])

# SB Rn,disp16(Rm) - 1100_nnnn_mmmm_1000 dddd_dddd_dddd_dddd
addop("SB", [bs("1100"), reg04, reg04_deref_noarg, bs("1000"), disp16_reg_deref], [reg04, disp16_reg_deref])

# SH Rn,disp16(Rm) - 1100_nnnn_mmmm_1001 dddd_dddd_dddd_dddd
addop("SH", [bs("1100"), reg04, reg04_deref_noarg, bs("1001"), disp16_reg_deref], [reg04, disp16_reg_deref])

# SW Rn,disp16(Rm) - 1100_nnnn_mmmm_1010 dddd_dddd_dddd_dddd
addop("SW", [bs("1100"), reg04, reg04_deref_noarg, bs("1010"), disp16_reg_deref], [reg04, disp16_reg_deref])

# LBU Rn,disp16(Rm) - 1100_nnnn_mmmm_1011 dddd_dddd_dddd_dddd
addop("LBU", [bs("1100"), reg04, reg04_deref_noarg, bs("1011"), disp16_reg_deref], [reg04, disp16_reg_deref])

# LB Rn,disp16(Rm) - 1100_nnnn_mmmm_1100 dddd_dddd_dddd_dddd
addop("LB", [bs("1100"), reg04, reg04_deref_noarg, bs("1100"), disp16_reg_deref], [reg04, disp16_reg_deref])

# LH Rn,disp16(Rm) - 1100_nnnn_mmmm_1101 dddd_dddd_dddd_dddd
addop("LH", [bs("1100"), reg04, reg04_deref_noarg, bs("1101"), disp16_reg_deref], [reg04, disp16_reg_deref])

# LW Rn,disp16(Rm) - 1100_nnnn_mmmm_1110 dddd_dddd_dddd_dddd
addop("LW", [bs("1100"), reg04, reg04_deref_noarg, bs("1110"), disp16_reg_deref], [reg04, disp16_reg_deref])

# LHU Rn,disp16(Rm) - 1100_nnnn_mmmm_1111 dddd_dddd_dddd_dddd
addop("LHU", [bs("1100"), reg04, reg04_deref_noarg, bs("1111"), disp16_reg_deref], [reg04, disp16_reg_deref])


### <Major Opcode #13>

# MOVU Rn[0-7],imm24 - 1101_0nnn_IIII_IIII iiii_iiii_iiii_iiii
addop("MOVU", [bs("11010"), reg03, imm24])

# BCPEQ cccc,disp17 - 1101_1000_cccc_0100 dddd_dddd_dddd_dddd
addop("BCPEQ", [bs("11011000"), imm4, bs("0100"), disp17])

# BCPNE cccc,disp17 - 1101_1000_cccc_0101 dddd_dddd_dddd_dddd
addop("BCPNE", [bs("11011000"), imm4, bs("0101"), disp17])

# BCPAT cccc,disp17 - 1101_1000_cccc_0110 dddd_dddd_dddd_dddd
addop("BCPAT", [bs("11011000"), imm4, bs("0110"), disp17])

# BCPAF cccc,disp17 - 1101_1000_cccc_0111 dddd_dddd_dddd_dddd
addop("BCPAF", [bs("11011000"), imm4, bs("0111"), disp17])

# JMP target24 - 1101_1TTT_TTTT_1000 tttt_tttt_tttt_tttt
addop("JMP", [bs("11011"), imm7_noarg, bs("1000"), imm_target24],
      [imm_target24])  # the only interesting operand is imm_target24

# BSR disp24 - 1101_1DDD_DDDD_1001 dddd_dddd_dddd_dddd
addop("BSR", [bs("11011"), imm7_noarg, bs("1001"), imm_target24_signed], [imm_target24_signed])

# BSRV disp24 1101_1DDD_DDDD_1011 dddd_dddd_dddd_dddd
addop("BSRV", [bs("11011"), imm7_noarg, bs("1011"), imm_target24], [imm_target24])


### <Major Opcode #14>

# BEQI Rn,imm4,disp17 - 1110_nnnn_iiii_0000 dddd_dddd_dddd_dddd
addop("BEQI", [bs("1110"), reg04, imm4, bs("0000"), disp17])

# BEQ Rn,Rm,disp17 - 1110_nnnn_mmmm_0001 dddd_dddd_dddd_dddd
addop("BEQ", [bs("1110"), reg04, reg04, bs("0001"), disp17])

# BNEI Rn,imm4,disp17 - 1110_nnnn_iiii_0100 dddd_dddd_dddd_dddd
addop("BNEI", [bs("1110"), reg04, imm4, bs("0100"), disp17])

# BNE Rn,Rm,disp17 - 1110_nnnn_mmmm_0101 dddd_dddd_dddd_dddd
addop("BNE", [bs("1110"), reg04, reg04, bs("0101"), disp17])

# BGEI Rn,imm4,disp17 - 1110_nnnn_iiii_1000 dddd_dddd_dddd_dddd
addop("BGEI", [bs("1110"), reg04, imm4, bs("1000"), disp17])

# REPEAT Rn,disp17 - 1110_nnnn_0000_1001 dddd_dddd_dddd_dddd
addop("REPEAT", [bs("1110"), reg04, bs("00001001"), disp17])

# EREPEAT disp17 - 1110_0000_0001_1001 dddd_dddd_dddd_dddd
addop("EREPEAT", [bs("1110000000011001"), disp17])

# BLTI Rn,imm4,disp17 - 1110_nnnn_iiii_1100 dddd_dddd_dddd_dddd
addop("BLTI", [bs("1110"), reg04, imm4, bs("1100"), disp17])

# (RI) - 1110_xxxx_xxxx_1101 xxxx_xxxx_xxxx_xxxx
addop("(RI)", [bs("1110"), imm8, bs("1101"), imm16])

# SW Rn,(abs24) - 1110_nnnn_DDDD_DD10 dddd_dddd_dddd_dddd
addop("SW", [bs("1110"), reg04, imm6_noarg, bs("10"), abs24])

# LW Rn,(abs24) - 1110_nnnn_DDDD_DD11 dddd_dddd_dddd_dddd
addop("LW", [bs("1110"), reg04, imm6_noarg, bs("11"), abs24])


### <Major Opcode #15>

# DSP Rn,Rm,code16 - 1111_nnnn_mmmm_0000 cccc_cccc_cccc_cccc
addop("DSP", [bs("1111"), reg04, reg04, bs("0000"), imm16])

# Note: DSP, DSP0 & DSP1 look exactly the same. This is ambiguous, and prevent
#       them for being correctly disassembled. DSP0 & DSP1 are arbitrarily
#       disabled.

# DSP0 code24 - 1111_nnnn_mmmm_0000 cccc_cccc_cccc_cccc
#addop("DSP0", [bs("1111"), imm8_noarg, bs("0000"), imm_code24], [imm_code24])

# DSP1 Rn,code20 - 1111_nnnn_mmmm_0000 cccc_cccc_cccc_cccc
#addop("DSP1", [bs("1111"), reg04, imm4_noarg, bs("0000"), imm_code20])

# LDZ Rn,Rm - 1111_nnnn_mmmm_0001 0000_0000_0000_0000
addop("LDZ", [bs("1111"), reg04, reg04, bs("00010000000000000000")])

# AVE Rn,Rm - 1111_nnnn_mmmm_0001 0000_0000_0000_0010
addop("AVE", [bs("1111"), reg04, reg04, bs("00010000000000000010")])

# ABS Rn,Rm - 1111_nnnn_mmmm_0001 0000_0000_0000_0011
addop("ABS", [bs("1111"), reg04, reg04, bs("00010000000000000011")])

# MIN Rn,Rm - 1111_nnnn_mmmm_0001 0000_0000_0000_0100
addop("MIN", [bs("1111"), reg04, reg04, bs("00010000000000000100")])

# MAX Rn,Rm - 1111_nnnn_mmmm_0001 0000_0000_0000_0101
addop("MAX", [bs("1111"), reg04, reg04, bs("00010000000000000101")])

# MINU Rn,Rm - 1111_nnnn_mmmm_0001 0000_0000_0000_0110
addop("MINU", [bs("1111"), reg04, reg04, bs("00010000000000000110")])

# MAXU Rn,Rm - 1111_nnnn_mmmm_0001 0000_0000_0000_0111
addop("MAXU", [bs("1111"), reg04, reg04, bs("00010000000000000111")])

# SADD Rn,Rm - 1111_nnnn_mmmm_0001 0000_0000_0000_1000
addop("SADD", [bs("1111"), reg04, reg04, bs("00010000000000001000")])

# SADDU Rn,Rm - 1111_nnnn_mmmm_0001 0000_0000_0000_1001
addop("SADDU", [bs("1111"), reg04, reg04, bs("00010000000000001001")])

# SSUB Rn,Rm - 1111_nnnn_mmmm_0001 0000_0000_0000_1010
addop("SSUB", [bs("1111"), reg04, reg04, bs("00010000000000001010")])

# SSUBU Rn,Rm - 1111_nnnn_mmmm_0001 0000_0000_0000_1011
addop("SSUBU", [bs("1111"), reg04, reg04, bs("00010000000000001011")])

# CLIP Rn,imm5 - 1111_nnnn_0000_0001 0001_0000_iiii_i000
addop("CLIP", [bs("1111"), reg04, bs("0000000100010000"), imm5, bs("000")])

# CLIPU Rn,imm5 - 1111_nnnn_0000_0001 0001_0000_iiii_i001
addop("CLIPU", [bs("1111"), reg04, bs("0000000100010000"), imm5, bs("001")])

# (RI) - 1111_xxxx_xxxx_0001 0010_xxxx_xxxx_xxxx
addop("(RI)", [bs("1111"), imm8, bs("00010010"), imm12])

# MADD Rn,Rm - 1111_nnnn_mmmm_0001 0011_0000_0000_0100
addop("MADD", [bs("1111"), reg04, reg04, bs("00010011000000000100")])

# MADDU Rn,Rm - 1111_nnnn_mmmm_0001 0011_0000_0000_0101
addop("MADDU", [bs("1111"), reg04, reg04, bs("00010011000000000101")])

# MADDR Rn,Rm - 1111_nnnn_mmmm_0001 0011_0000_0000_0110
addop("MADDR", [bs("1111"), reg04, reg04, bs("00010011000000000110")])

# MADDRU Rn,Rm - 1111_nnnn_mmmm_0001 0011_0000_0000_0111
addop("MADDRU", [bs("1111"), reg04, reg04, bs("00010011000000000111")])

# UCI Rn,Rm,code16 - 1111_nnnn_mmmm_0010 cccc_cccc_cccc_cccc
addop("UCI", [bs("1111"), reg04, reg04, bs("0010"), imm16])

# (RI) - 1111_xxxx_xxxx_0011 xxxx_xxxx_xxxx_xxxx
addop("(RI)", [bs("1111"), imm8, bs("0011"), imm16])

# STCB Rn,abs16 - 1111_nnnn_0000_0100 aaaa_aaaa_aaaa_aaaa
addop("STCB", [bs("1111"), reg04, bs("00000100"), imm16])

# LDCB Rn,abs16 - 1111_nnnn_0001_0100 aaaa_aaaa_aaaa_aaaa
addop("LDCB", [bs("1111"), reg04, bs("00010100"), imm16])

# SBCPA CRn,(Rm+),imm8 - 1111_nnnn_mmmm_0101 0000_0000_iiii_iiii
addop("SBCPA", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010100000000"), imm8])

# SHCPA CRn,(Rm+),imm8.align2 - 1111_nnnn_mmmm_0101 0001_0000_iiii_iii0
addop("SHCPA", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010100010000"), imm8_align2, bs("0")])

# SWCPA CRn,(Rm+),imm8.align4 - 1111_nnnn_mmmm_0101 0010_0000_iiii_ii00
addop("SWCPA", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010100100000"), imm8_align4, bs("00")])

# SMCPA CRn,(Rm+),imm8.align8 - 1111_nnnn_mmmm_0101 0011_0000_iiii_i000
addop("SMCPA", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010100110000"), imm8_align8, bs("000")])

# LBCPA CRn,(Rm+),imm8 - 1111_nnnn_mmmm_0101 0100_0000_iiii_iiii
addop("LBCPA", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010101000000"), imm8])

# LHCPA CRn,(Rm+),imm8.align2 - 1111_nnnn_mmmm_0101 0101_0000_iiii_iii0
addop("LHCPA", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010101010000"), imm8_align2, bs("0")])

# LWCPA CRn,(Rm+),imm8.align4 - 1111_nnnn_mmmm_0101 0110_0000_iiii_ii00
addop("LWCPA", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010101100000"), imm8_align4, bs("00")])

# LMCPA CRn,(Rm+),imm8.align8 - 1111_nnnn_mmmm_0101 0111_0000_iiii_i000
addop("LMCPA", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010101110000"), imm8_align8, bs("000")])

# SBCPM0 CRn,(Rm+),imm8 - 1111_nnnn_mmmm_0101 0000_1000_iiii_iiii
addop("SBCPM0", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010100001000"), imm8])

# SHCPM0 CRn,(Rm+),imm8.align2 - 1111_nnnn_mmmm_0101 0001_1000_iiii_iii0
addop("SHCPM0", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010100011000"), imm8_align2, bs("0")])

# SWCPM0 CRn,(Rm+),imm8.align4 - 1111_nnnn_mmmm_0101 0010_1000_iiii_ii00
addop("SWCPM0", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010100101000"), imm8_align4, bs("00")])

# SMCPM0 CRn,(Rm+),imm8.align8 - 1111_nnnn_mmmm_0101 0011_1000_iiii_i000
addop("SMCPM0", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010100111000"), imm8_align8, bs("000")])

# LBCPM0 CRn,(Rm+),imm8 - 1111_nnnn_mmmm_0101 0100_1000_iiii_iiii
addop("LBCPM0", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010101001000"), imm8])

# LHCPM0 CRn,(Rm+),imm8.align2 - 1111_nnnn_mmmm_0101 0101_1000_iiii_iii0
addop("LHCPM0", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010101011000"), imm8_align2, bs("0")])

# LWCPM0 CRn,(Rm+),imm8.align4 - 1111_nnnn_mmmm_0101 0110_1000_iiii_ii00
addop("LWCPM0", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010101101000"), imm8_align4, bs("00")])

# LMCPM0 CRn,(Rm+),imm8.align8 - 1111_nnnn_mmmm_0101 0111_1000_iiii_i000
addop("LMCPM0", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010101111000"), imm8_align8, bs("000")])

# SBCPM1 CRn,(Rm+),imm8 - 1111_nnnn_mmmm_0101 0000_1100_iiii_iiii
addop("SBCPM1", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010100001100"), imm8])

# SHCPM1 CRn,(Rm+),imm8.align2 - 1111_nnnn_mmmm_0101 0001_1100_iiii_iii0
addop("SHCPM1", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010100011100"), imm8_align2, bs("0")])

# SWCPM1 CRn,(Rm+),imm8.align4 - 1111_nnnn_mmmm_0101 0010_1100_iiii_ii00
addop("SWCPM1", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010100101100"), imm8_align4, bs("00")])

# SMCPM1 CRn,(Rm+),imm8.align8 - 1111_nnnn_mmmm_0101 0011_1100_iiii_i000
addop("SMCPM1", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010100111100"), imm8_align8, bs("000")])

# LBCPM1 CRn,(Rm+),imm8 - 1111_nnnn_mmmm_0101 0100_1100_iiii_iiii
addop("LBCPM1", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010101001100"), imm8])

# LHCPM1 CRn,(Rm+),imm8.align2 - 1111_nnnn_mmmm_0101 0101_1100_iiii_iii0
addop("LHCPM1", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010101011100"), imm8_align2, bs("0")])

# LWCPM1 CRn,(Rm+),imm8.align4 - 1111_nnnn_mmmm_0101 0110_1100_iiii_ii00
addop("LWCPM1", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010101101100"), imm8_align4, bs("00")])

# LMCPM1 CRn,(Rm+),imm8.align8 - 1111_nnnn_mmmm_0101 0111_1100_iiii_i000
addop("LMCPM1", [bs("1111"), copro_reg04, reg04_inc_deref, bs("010101111100"), imm8_align8, bs("000")])

# (RI) - 1111_xxxx_xxxx_0110 xxxx_xxxx_xxxx_xxxx
addop("(RI)", [bs("1111"), imm8, bs("0110"), imm16])

# CP code24 - 1111_CCCC_CCCC_0111 cccc_cccc_cccc_cccc
#addop("CP", [bs("1111"), imm8_noarg, bs("0111"), imm_code24], [imm_code24])
# Note: CP & CMOV* look exactly the same. This is ambiguous, and prevent
#       them for being correctly disassembled. CP was arbitrarily disabled.

# CP code56 - 1111_CCCC_CCCC_0111 cccc_cccc_cccc_cccc cccc_cccc_cccc_cccc
# 64-bit VLIW operation mode - not implemented

# CMOV CRn,Rm - 1111_nnnn_mmmm_0111 1111_0000_0000_0000
#addop("CMOV", [bs("1111"), copro_reg04, reg04, bs("01111111000000000000")])

# CMOV Rm,CRn - 1111_nnnn_mmmm_0111 1111_0000_0000_0001
#addop("CMOV", [bs("1111"), copro_reg04, reg04, bs("01111111000000000001")], [reg04, copro_reg04])

# CMOVC CCRn,Rm - 1111_nnnn_mmmm_0111 1111_0000_0000_NN10
# CRn=NNnnnn
addop("CMOVC", [bs("1111"), imm4_noarg, reg04, bs("0111111100000000"), copro_reg06, bs("10")], [copro_reg06, reg04])

# CMOVC Rm,CCRn - 1111_nnnn_mmmm_0111 1111_0000_0000_NN11
# CRn=NNnnnn
addop("CMOVC", [bs("1111"), imm4_noarg, reg04, bs("0111111100000000"), copro_reg06, bs("11")], [reg04, copro_reg06])

# CMOVH CRn,Rm - 1111_nnnn_mmmm_0111 1111_0001_0000_0000
#addop("CMOVH", [bs("1111"), copro_reg04, reg04, bs("01111111000100000000")])

# CMOVH Rm,CRn - 1111_nnnn_mmmm_0111 1111_0001_0000_0001
#addop("CMOVH", [bs("1111"), copro_reg04, reg04, bs("01111111000100000001")], [reg04, copro_reg04])

# Note: the following CMOV* instructions are extensions used when the processor
#       has more than 16 coprocessor general-purpose registers. They can be
#       used to assemble and disassemble both CMOV* instructuons sets.

# CMOV CRn,Rm - 1111_nnnn_mmmm_0111 1111_0000_0000_N000
# CRn=Nnnnn
addop("CMOV", [bs("1111"), imm4_noarg, reg04, bs("0111111100000000"), copro_reg05, bs("000")], [copro_reg05, reg04])

# CMOV Rm,CRn - 1111_nnnn_mmmm_0111 1111_0000_0000_N001
addop("CMOV", [bs("1111"), imm4_noarg, reg04, bs("0111111100000000"), copro_reg05, bs("001")], [reg04, copro_reg05])

# CMOVH CRn,Rm - 1111_nnnn_mmmm_0111 1111_0001_0000_N000
addop("CMOVH", [bs("1111"), imm4_noarg, reg04, bs("0111111100010000"), copro_reg05, bs("000")], [copro_reg05, reg04])

# CMOVH Rm,CRn - 1111_nnnn_mmmm_0111 1111_0001_0000_N001
addop("CMOVH", [bs("1111"), imm4_noarg, reg04, bs("0111111100010000"), copro_reg05, bs("001")], [reg04, copro_reg05])

# (RI) - 1111_xxxx_xxxx_10xx xxxx_xxxx_xxxx_xxxx
addop("(RI)", [bs("1111"), imm8, bs("10"), imm18])

# SWCP CRn,disp16(Rm) - 1111_nnnn_mmmm_1100 dddd_dddd_dddd_dddd
addop("SWCP", [bs("1111"), copro_reg04, reg04_deref_noarg, bs("1100"), disp16_reg_deref], [copro_reg04, disp16_reg_deref])

# LWCP CRn,disp16(Rm) - 1111_nnnn_mmmm_1101 dddd_dddd_dddd_dddd
addop("LWCP", [bs("1111"), copro_reg04, reg04_deref_noarg, bs("1101"), disp16_reg_deref], [copro_reg04, disp16_reg_deref, reg04_deref])

# SMCP CRn,disp16(Rm) - 1111_nnnn_mmmm_1110 dddd_dddd_dddd_dddd
addop("SMCP", [bs("1111"), copro_reg04, reg04_deref_noarg, bs("1110"), disp16_reg_deref], [copro_reg04, disp16_reg_deref, reg04_deref])

# LMCP CRn,disp16(Rm) - 1111_nnnn_mmmm_1111 dddd_dddd_dddd_dddd
addop("LMCP", [bs("1111"), copro_reg04, reg04_deref_noarg, bs("1111"), disp16_reg_deref], [copro_reg04, disp16_reg_deref])



from builtins import map
from builtins import range
import imp
import logging

# Raise an ImportError if z3 is not available WITHOUT actually importing it
imp.find_module("z3")

from miasm.ir.translators.translator import Translator

log = logging.getLogger("translator_z3")
console_handler = logging.StreamHandler()
console_handler.setFormatter(logging.Formatter("[%(levelname)-8s]: %(message)s"))
log.addHandler(console_handler)
log.setLevel(logging.WARNING)

class Z3Mem(object):
    """Memory abstraction for TranslatorZ3. Memory elements are only accessed,
    never written. To give a concrete value for a given memory cell in a solver,
    add "mem32.get(address, size) == <value>" constraints to your equation.
    The endianness of memory accesses is handled accordingly to the "endianness"
    attribute.

    Note: Will have one memory space for each addressing size used.
    For example, if memory is accessed via 32 bits values and 16 bits values,
    these access will not occur in the same address space.
    """

    def __init__(self, endianness="<", name="mem"):
        """Initializes a Z3Mem object with a given @name and @endianness.
        @endianness: Endianness of memory representation. '<' for little endian,
            '>' for big endian.
        @name: name of memory Arrays generated. They will be named
            name+str(address size) (for example mem32, mem16...).
        """
        # Import z3 only on demand
        global z3
        import z3

        if endianness not in ['<', '>']:
            raise ValueError("Endianness should be '>' (big) or '<' (little)")
        self.endianness = endianness
        self.mems = {} # Address size -> memory z3.Array
        self.name = name

    def get_mem_array(self, size):
        """Returns a z3 Array used internally to represent memory for addresses
        of size @size.
        @size: integer, size in bit of addresses in the memory to get.
        Return a z3 Array: BitVecSort(size) -> BitVecSort(8).
        """
        try:
            mem = self.mems[size]
        except KeyError:
            # Lazy instantiation
            self.mems[size] = z3.Array(self.name + str(size),
                                        z3.BitVecSort(size),
                                        z3.BitVecSort(8))
            mem = self.mems[size]
        return mem

    def __getitem__(self, addr):
        """One byte memory access. Different address sizes with the same value
        will result in different memory accesses.
        @addr: a z3 BitVec, the address to read.
        Return a z3 BitVec of size 8 bits representing a memory access.
        """
        size = addr.size()
        mem = self.get_mem_array(size)
        return mem[addr]

    def get(self, addr, size):
        """ Memory access at address @addr of size @size.
        @addr: a z3 BitVec, the address to read.
        @size: int, size of the read in bits.
        Return a z3 BitVec of size @size representing a memory access.
        """
        original_size = size
        if original_size % 8 != 0:
            # Size not aligned on 8bits -> read more than size and extract after
            size = ((original_size // 8) + 1) * 8
        res = self[addr]
        if self.is_little_endian():
            for i in range(1, size // 8):
                res = z3.Concat(self[addr+i], res)
        else:
            for i in range(1, size //8):
                res = z3.Concat(res, self[addr+i])
        if size == original_size:
            return res
        else:
            # Size not aligned, extract right sized result
            return z3.Extract(original_size-1, 0, res)

    def is_little_endian(self):
        """True if this memory is little endian."""
        return self.endianness == "<"

    def is_big_endian(self):
        """True if this memory is big endian."""
        return not self.is_little_endian()


class TranslatorZ3(Translator):
    """Translate a Miasm expression to an equivalent z3 python binding
    expression. Memory is abstracted via z3.Array (see Z3Mem).
    The result of from_expr will be a z3 Expr.

    If you want to interact with the memory abstraction after the translation,
    you can instantiate your own Z3Mem, that will be equivalent to the one
    used by TranslatorZ3.
    """

    # Implemented language
    __LANG__ = "z3"
    # Operations translation
    trivial_ops = ["+", "-", "/", "%", "&", "^", "|", "*", "<<"]

    def __init__(self, endianness="<", loc_db=None, **kwargs):
        """Instance a Z3 translator
        @endianness: (optional) memory endianness
        """
        # Import z3 only on demand
        global z3
        import z3

        super(TranslatorZ3, self).__init__(**kwargs)
        self._mem = Z3Mem(endianness)
        self.loc_db = loc_db

    def from_ExprInt(self, expr):
        return z3.BitVecVal(int(expr), expr.size)

    def from_ExprId(self, expr):
        return z3.BitVec(str(expr), expr.size)

    def from_ExprLoc(self, expr):
        if self.loc_db is None:
            # No loc_db, fallback to default name
            return z3.BitVec(str(expr), expr.size)
        loc_key = expr.loc_key
        offset = self.loc_db.get_location_offset(loc_key)
        if offset is not None:
            return z3.BitVecVal(offset, expr.size)
        # fallback to default name
        return z3.BitVec(str(loc_key), expr.size)

    def from_ExprMem(self, expr):
        addr = self.from_expr(expr.ptr)
        return self._mem.get(addr, expr.size)

    def from_ExprSlice(self, expr):
        res = self.from_expr(expr.arg)
        res = z3.Extract(expr.stop-1, expr.start, res)
        return res

    def from_ExprCompose(self, expr):
        res = None
        for arg in expr.args:
            e = z3.Extract(arg.size-1, 0, self.from_expr(arg))
            if res != None:
                res = z3.Concat(e, res)
            else:
                res = e
        return res

    def from_ExprCond(self, expr):
        cond = self.from_expr(expr.cond)
        src1 = self.from_expr(expr.src1)
        src2 = self.from_expr(expr.src2)
        return z3.If(cond != 0, src1, src2)

    def _abs(self, z3_value):
        return z3.If(z3_value >= 0,z3_value,-z3_value)

    def _sdivC(self, num_expr, den_expr):
        """Divide (signed) @num by @den (Expr) as C would
        See modint.__div__ for implementation choice
        """
        num, den = self.from_expr(num_expr), self.from_expr(den_expr)
        num_s = self.from_expr(num_expr.signExtend(num_expr.size * 2))
        den_s = self.from_expr(den_expr.signExtend(den_expr.size * 2))
        result_sign = z3.If(num_s * den_s >= 0,
                            z3.BitVecVal(1, num.size()),
                            z3.BitVecVal(-1, num.size()),
        )
        return z3.UDiv(self._abs(num), self._abs(den)) * result_sign

    def from_ExprOp(self, expr):
        args = list(map(self.from_expr, expr.args))
        res = args[0]

        if len(args) > 1:
            for arg in args[1:]:
                if expr.op in self.trivial_ops:
                    res = eval("res %s arg" % expr.op)
                elif expr.op == ">>":
                    res = z3.LShR(res, arg)
                elif expr.op == "a>>":
                    res = res >> arg
                elif expr.op == "<<<":
                    res = z3.RotateLeft(res, arg)
                elif expr.op == ">>>":
                    res = z3.RotateRight(res, arg)
                elif expr.op == "sdiv":
                    res = self._sdivC(expr.args[0], expr.args[1])
                elif expr.op == "udiv":
                    res = z3.UDiv(res, arg)
                elif expr.op == "smod":
                    res = res - (arg * (self._sdivC(expr.args[0], expr.args[1])))
                elif expr.op == "umod":
                    res = z3.URem(res, arg)
                elif expr.op == "==":
                    res = z3.If(
                        args[0] == args[1],
                        z3.BitVecVal(1, 1),
                        z3.BitVecVal(0, 1)
                    )
                elif expr.op == "<u":
                    res = z3.If(
                        z3.ULT(args[0], args[1]),
                        z3.BitVecVal(1, 1),
                        z3.BitVecVal(0, 1)
                    )
                elif expr.op == "<s":
                    res = z3.If(
                        args[0] < args[1],
                        z3.BitVecVal(1, 1),
                        z3.BitVecVal(0, 1)
                    )
                elif expr.op == "<=u":
                    res = z3.If(
                        z3.ULE(args[0], args[1]),
                        z3.BitVecVal(1, 1),
                        z3.BitVecVal(0, 1)
                    )
                elif expr.op == "<=s":
                    res = z3.If(
                        args[0] <= args[1],
                        z3.BitVecVal(1, 1),
                        z3.BitVecVal(0, 1)
                    )
                else:
                    raise NotImplementedError("Unsupported OP yet: %s" % expr.op)
        elif expr.op == 'parity':
            arg = z3.Extract(7, 0, res)
            res = z3.BitVecVal(1, 1)
            for i in range(8):
                res = res ^ z3.Extract(i, i, arg)
        elif expr.op == '-':
            res = -res
        elif expr.op == "cnttrailzeros":
            size = expr.size
            src = res
            res = z3.If(src == 0, size, src)
            for i in range(size - 1, -1, -1):
                res = z3.If((src & (1 << i)) != 0, i, res)
        elif expr.op == "cntleadzeros":
            size = expr.size
            src = res
            res = z3.If(src == 0, size, src)
            for i in range(size, 0, -1):
                index = - i % size
                out = size - (index + 1)
                res = z3.If((src & (1 << index)) != 0, out, res)
        elif expr.op.startswith("zeroExt"):
            arg, = expr.args
            res = z3.ZeroExt(expr.size - arg.size, self.from_expr(arg))
        elif expr.op.startswith("signExt"):
            arg, = expr.args
            res = z3.SignExt(expr.size - arg.size, self.from_expr(arg))
        else:
            raise NotImplementedError("Unsupported OP yet: %s" % expr.op)

        return res

    def from_ExprAssign(self, expr):
        src = self.from_expr(expr.src)
        dst = self.from_expr(expr.dst)
        return (src == dst)


# Register the class
Translator.register(TranslatorZ3)



from __future__ import print_function
import os
import tempfile
from builtins import int as int_types

from future.utils import viewitems, viewvalues

import idaapi
import ida_kernwin
import idc
import ida_funcs
import idautils

from miasm.core.bin_stream_ida import bin_stream_ida
from miasm.expression.simplifications import expr_simp
from miasm.ir.ir import IRBlock, AssignBlock
from miasm.analysis.data_flow import load_from_int
from utils import guess_machine, expr2colorstr
from miasm.expression.expression import ExprLoc, ExprInt, ExprOp, ExprAssign
from miasm.analysis.simplifier import IRCFGSimplifierCommon, IRCFGSimplifierSSA
from miasm.core.locationdb import LocationDB



TYPE_GRAPH_IR = 0
TYPE_GRAPH_IRSSA = 1
TYPE_GRAPH_IRSSAUNSSA = 2

OPTION_GRAPH_CODESIMPLIFY = 1
OPTION_GRAPH_USE_IDA_STACK = 2
OPTION_GRAPH_DONTMODSTACK = 4
OPTION_GRAPH_LOADMEMINT = 8


class GraphIRForm(ida_kernwin.Form):

    def __init__(self):
        ida_kernwin.Form.__init__(
            self,
            r"""BUTTON YES* Launch
BUTTON CANCEL NONE
Graph IR Settings

{FormChangeCb}
Analysis:
<Graph IR :{rGraphIR}>
<Graph IR + SSA :{rGraphIRSSA}>
<Graph IR + SSA + UnSSA :{rGraphIRSSAUNSSA}>{cScope}>

Options:
<Simplify code:{rCodeSimplify}>
<Use ida stack:{rUseIdaStack}>
<Subcalls dont change stack:{rDontModStack}>
<Load static memory:{rLoadMemInt}>{cOptions}>
""",
            {
                'FormChangeCb': ida_kernwin.Form.FormChangeCb(self.OnFormChange),
                'cScope': ida_kernwin.Form.RadGroupControl(
                    (
                        "rGraphIR",
                        "rGraphIRSSA",
                        "rGraphIRSSAUNSSA"
                    )
                ),
                'cOptions': ida_kernwin.Form.ChkGroupControl(
                    (
                        "rCodeSimplify",
                        "rUseIdaStack",
                        "rDontModStack",
                        "rLoadMemInt"
                    )
                ),
            }
        )
        form, _ = self.Compile()
        form.rCodeSimplify.checked = True
        form.rUseIdaStack.checked = True
        form.rDontModStack.checked = False
        form.rLoadMemInt.checked = False

    def OnFormChange(self, _):
        return 1


# Override Miasm asmblock default label naming convention to shrink block size
# in IDA

def label_init(self, name="", offset=None):
    self.fixedblocs = False
    if isinstance(name, int_types):
        name = "loc_%X" % (int(name) & 0xFFFFFFFFFFFFFFFF)
    self.name = name
    self.attrib = None
    if offset is None:
        self.offset = None
    else:
        self.offset = int(offset)


def label_str(self):
    if isinstance(self.offset, int_types):
        return "%s:0x%x" % (self.name, self.offset)
    return "%s:%s" % (self.name, self.offset)


def color_irblock(irblock, lifter):
    out = []
    lbl = idaapi.COLSTR("%s:" % lifter.loc_db.pretty_str(irblock.loc_key), idaapi.SCOLOR_INSN)
    out.append(lbl)
    for assignblk in irblock:
        for dst, src in sorted(viewitems(assignblk)):
            dst_f = expr2colorstr(dst, loc_db=lifter.loc_db)
            src_f = expr2colorstr(src, loc_db=lifter.loc_db)
            line = idaapi.COLSTR("%s = %s" % (dst_f, src_f), idaapi.SCOLOR_INSN)
            out.append('    %s' % line)
        out.append("")
    out.pop()
    return "\n".join(out)


class GraphMiasmIR(idaapi.GraphViewer):

    def __init__(self, ircfg, title, result):
        idaapi.GraphViewer.__init__(self, title)
        self.ircfg = ircfg
        self.result = result
        self.names = {}

    def OnRefresh(self):
        self.Clear()
        addr_id = {}
        for (loc_key, irblock) in viewitems(self.ircfg.blocks):
            id_irblock = self.AddNode(color_irblock(irblock, self.ircfg))
            addr_id[loc_key] = id_irblock

        for (loc_key, irblock) in viewitems(self.ircfg.blocks):
            if not irblock:
                continue
            all_dst = self.ircfg.dst_trackback(irblock)
            for dst in all_dst:
                if not dst.is_loc():
                    continue
                if not dst.loc_key in self.ircfg.blocks:
                    continue
                node1 = addr_id[loc_key]
                node2 = addr_id[dst.loc_key]
                self.AddEdge(node1, node2)
        return True

    def OnGetText(self, node_id):
        return str(self[node_id])

    def OnSelect(self, _):
        return True

    def OnClick(self, _):
        return True

    def Show(self):
        if not idaapi.GraphViewer.Show(self):
            return False
        return True


def is_addr_ro_variable(bs, addr, size):
    """
    Return True if address at @addr is a read-only variable.
    WARNING: Quick & Dirty

    @addr: integer representing the address of the variable
    @size: size in bits

    """
    try:
        _ = bs.getbytes(addr, size // 8)
    except IOError:
        return False
    return True


def build_graph(start_addr, type_graph, simplify=False, use_ida_stack=True, dontmodstack=False, loadint=False, verbose=False):
    machine = guess_machine(addr=start_addr)
    dis_engine, lifter_model_call = machine.dis_engine, machine.lifter_model_call

    class IRADelModCallStack(lifter_model_call):
        def call_effects(self, addr, instr):
            assignblks, extra = super(IRADelModCallStack, self).call_effects(addr, instr)
            if use_ida_stack:
                stk_before = idc.get_spd(instr.offset)
                stk_after = idc.get_spd(instr.offset + instr.l)
                stk_diff = stk_after - stk_before
                print(hex(stk_diff))
                call_assignblk = AssignBlock(
                    [
                        ExprAssign(self.ret_reg, ExprOp('call_func_ret', addr)),
                        ExprAssign(self.sp, self.sp + ExprInt(stk_diff, self.sp.size))
                    ],
                    instr
                )
                return [call_assignblk], []
            else:
                if not dontmodstack:
                    return assignblks, extra
                out = []
                for assignblk in assignblks:
                    dct = dict(assignblk)
                    dct = {
                        dst:src for (dst, src) in viewitems(dct) if dst != self.sp
                    }
                    out.append(AssignBlock(dct, assignblk.instr))
            return out, extra


    if verbose:
        print("Arch", dis_engine)

    fname = idc.get_root_filename()
    if verbose:
        print(fname)

    bs = bin_stream_ida()
    loc_db = LocationDB()

    mdis = dis_engine(bs, loc_db=loc_db)
    lifter = IRADelModCallStack(loc_db)


    # populate symbols with ida names
    for addr, name in idautils.Names():
        if name is None:
            continue
        if (loc_db.get_offset_location(addr) or
            loc_db.get_name_location(name)):
            # Symbol alias
            continue
        loc_db.add_location(name, addr)

    if verbose:
        print("start disasm")
    if verbose:
        print(hex(start_addr))

    asmcfg = mdis.dis_multiblock(start_addr)
    entry_points = set([loc_db.get_offset_location(start_addr)])
    if verbose:
        print("generating graph")
        open('asm_flow.dot', 'w').write(asmcfg.dot())
        print("generating IR... %x" % start_addr)

    ircfg = lifter.new_ircfg_from_asmcfg(asmcfg)

    if verbose:
        print("IR ok... %x" % start_addr)

    for irb in list(viewvalues(ircfg.blocks)):
        irs = []
        for assignblk in irb:
            new_assignblk = {
                expr_simp(dst): expr_simp(src)
                for dst, src in viewitems(assignblk)
            }
            irs.append(AssignBlock(new_assignblk, instr=assignblk.instr))
        ircfg.blocks[irb.loc_key] = IRBlock(loc_db, irb.loc_key, irs)

    if verbose:
        out = ircfg.dot()
        open(os.path.join(tempfile.gettempdir(), 'graph.dot'), 'wb').write(out)
    title = "Miasm IR graph"


    head = list(entry_points)[0]

    if simplify:
        ircfg_simplifier = IRCFGSimplifierCommon(lifter)
        ircfg_simplifier.simplify(ircfg, head)
        title += " (simplified)"

    if type_graph == TYPE_GRAPH_IR:
        graph = GraphMiasmIR(ircfg, title, None)
        graph.Show()
        return


    class IRAOutRegs(lifter_model_call):
        def get_out_regs(self, block):
            regs_todo = super(IRAOutRegs, self).get_out_regs(block)
            out = {}
            for assignblk in block:
                for dst in assignblk:
                    reg = self.ssa_var.get(dst, None)
                    if reg is None:
                        continue
                    if reg in regs_todo:
                        out[reg] = dst
            return set(viewvalues(out))



    # Add dummy dependency to uncover out regs affectation
    for loc in ircfg.leaves():
        irblock = ircfg.blocks.get(loc)
        if irblock is None:
            continue
        regs = {}
        for reg in lifter.get_out_regs(irblock):
            regs[reg] = reg
        assignblks = list(irblock)
        new_assiblk = AssignBlock(regs, assignblks[-1].instr)
        assignblks.append(new_assiblk)
        new_irblock = IRBlock(irblock.loc_db, irblock.loc_key, assignblks)
        ircfg.blocks[loc] = new_irblock


    class CustomIRCFGSimplifierSSA(IRCFGSimplifierSSA):
        def do_simplify(self, ssa, head):
            modified = super(CustomIRCFGSimplifierSSA, self).do_simplify(ssa, head)
            if loadint:
                modified |= load_from_int(ssa.graph, bs, is_addr_ro_variable)
            return modified

        def simplify(self, ircfg, head):
            ssa = self.ircfg_to_ssa(ircfg, head)
            ssa = self.do_simplify_loop(ssa, head)

            if type_graph == TYPE_GRAPH_IRSSA:
                ret = ssa.graph
            elif type_graph == TYPE_GRAPH_IRSSAUNSSA:
                ircfg = self.ssa_to_unssa(ssa, head)
                ircfg_simplifier = IRCFGSimplifierCommon(self.lifter)
                ircfg_simplifier.simplify(ircfg, head)
                ret = ircfg
            else:
                raise ValueError("Unknown option")
            return ret


    head = list(entry_points)[0]
    simplifier = CustomIRCFGSimplifierSSA(lifter)
    ircfg = simplifier.simplify(ircfg, head)
    open('final.dot', 'w').write(ircfg.dot())


    graph = GraphMiasmIR(ircfg, title, None)
    graph.Show()

def function_graph_ir():
    # Get settings
    settings = GraphIRForm()
    ret = settings.Execute()
    if not ret:
        return

    func = ida_funcs.get_func(idc.get_screen_ea())
    func_addr = func.start_ea

    build_graph(
        func_addr,
        settings.cScope.value,
        simplify=settings.cOptions.value & OPTION_GRAPH_CODESIMPLIFY,
        use_ida_stack=settings.cOptions.value & OPTION_GRAPH_USE_IDA_STACK,
        dontmodstack=settings.cOptions.value & OPTION_GRAPH_DONTMODSTACK,
        loadint=settings.cOptions.value & OPTION_GRAPH_LOADMEMINT,
        verbose=False
    )
    return

if __name__ == "__main__":
    function_graph_ir()



from future.utils import viewitems

import miasm.expression.expression as m2_expr
from miasm.core.utils import BoundedDict


class Translator(object):
    "Abstract parent class for translators."

    # Registered translators
    available_translators = []
    # Implemented language
    __LANG__ = ""

    @classmethod
    def register(cls, translator):
        """Register a translator
        @translator: Translator sub-class
        """
        cls.available_translators.append(translator)

    @classmethod
    def to_language(cls, target_lang, *args, **kwargs):
        """Return the corresponding translator instance
        @target_lang: str (case insensitive) wanted language
        Raise a NotImplementedError in case of unmatched language
        """
        target_lang = target_lang.lower()
        for translator in cls.available_translators:
            if translator.__LANG__.lower() == target_lang:
                return translator(*args, **kwargs)

        raise NotImplementedError("Unknown target language: %s" % target_lang)

    @classmethod
    def available_languages(cls):
        "Return the list of registered languages"
        return [translator.__LANG__ for translator in cls.available_translators]

    def __init__(self, cache_size=1000):
        """Instance a translator
        @cache_size: (optional) Expr cache size
        """
        self._cache = BoundedDict(cache_size)

    def from_ExprInt(self, expr):
        """Translate an ExprInt
        @expr: ExprInt to translate
        """
        raise NotImplementedError("Abstract method")

    def from_ExprId(self, expr):
        """Translate an ExprId
        @expr: ExprId to translate
        """
        raise NotImplementedError("Abstract method")

    def from_ExprLoc(self, expr):
        """Translate an ExprLoc
        @expr: ExprLoc to translate
        """
        raise NotImplementedError("Abstract method")

    def from_ExprCompose(self, expr):
        """Translate an ExprCompose
        @expr: ExprCompose to translate
        """
        raise NotImplementedError("Abstract method")

    def from_ExprSlice(self, expr):
        """Translate an ExprSlice
        @expr: ExprSlice to translate
        """
        raise NotImplementedError("Abstract method")

    def from_ExprOp(self, expr):
        """Translate an ExprOp
        @expr: ExprOp to translate
        """
        raise NotImplementedError("Abstract method")

    def from_ExprMem(self, expr):
        """Translate an ExprMem
        @expr: ExprMem to translate
        """
        raise NotImplementedError("Abstract method")

    def from_ExprAssign(self, expr):
        """Translate an ExprAssign
        @expr: ExprAssign to translate
        """
        raise NotImplementedError("Abstract method")

    def from_ExprCond(self, expr):
        """Translate an ExprCond
        @expr: ExprCond to translate
        """
        raise NotImplementedError("Abstract method")

    def from_expr(self, expr):
        """Translate an expression according to its type
        @expr: expression to translate
        """
        # Use cache
        if expr in self._cache:
            return self._cache[expr]

        # Handle Expr type
        handlers = {
            m2_expr.ExprInt: self.from_ExprInt,
            m2_expr.ExprId: self.from_ExprId,
            m2_expr.ExprLoc: self.from_ExprLoc,
            m2_expr.ExprCompose: self.from_ExprCompose,
            m2_expr.ExprSlice: self.from_ExprSlice,
            m2_expr.ExprOp: self.from_ExprOp,
            m2_expr.ExprMem: self.from_ExprMem,
            m2_expr.ExprAssign: self.from_ExprAssign,
            m2_expr.ExprCond: self.from_ExprCond
        }
        for target, handler in viewitems(handlers):
            if isinstance(expr, target):
                ## Compute value and update the internal cache
                ret = handler(expr)
                self._cache[expr] = ret
                return ret
        raise ValueError("Unhandled type for %s" % expr)




################################################################################
#
# By choice, Miasm2 does not handle comparison as a single operation, but with
# operations corresponding to comparison computation.
# One may want to detect those comparison; this library is designed to add them
# in Miasm2 engine thanks to :
# - Conditions computation in ExprOp
# - Simplifications to catch known condition forms
#
# Conditions currently supported :
# <u, <s, ==
#
# Authors : Fabrice DESCLAUX (CEA/DAM), Camille MOUGEY (CEA/DAM)
#
################################################################################

import miasm.expression.expression as m2_expr


# Jokers for expression matching

jok1 = m2_expr.ExprId("jok1", 32)
jok2 = m2_expr.ExprId("jok2", 32)
jok3 = m2_expr.ExprId("jok3", 32)
jok_small = m2_expr.ExprId("jok_small", 1)


# Constructors

def __ExprOp_cond(op, arg1, arg2):
    "Return an ExprOp standing for arg1 op arg2 with size to 1"
    ec = m2_expr.ExprOp(op, arg1, arg2)
    return ec


def ExprOp_inf_signed(arg1, arg2):
    "Return an ExprOp standing for arg1 <s arg2"
    return __ExprOp_cond(m2_expr.TOK_INF_SIGNED, arg1, arg2)


def ExprOp_inf_unsigned(arg1, arg2):
    "Return an ExprOp standing for arg1 <s arg2"
    return __ExprOp_cond(m2_expr.TOK_INF_UNSIGNED, arg1, arg2)

def ExprOp_equal(arg1, arg2):
    "Return an ExprOp standing for arg1 == arg2"
    return __ExprOp_cond(m2_expr.TOK_EQUAL, arg1, arg2)


# Catching conditions forms

def __check_msb(e):
    """If @e stand for the most significant bit of its arg, return the arg;
    False otherwise"""

    if not isinstance(e, m2_expr.ExprSlice):
        return False

    arg = e.arg
    if e.start != (arg.size - 1) or e.stop != arg.size:
        return False

    return arg

def __match_expr_wrap(e, to_match, jok_list):
    "Wrapper around match_expr to canonize pattern"

    to_match = to_match.canonize()

    r = m2_expr.match_expr(e, to_match, jok_list)
    if r is False:
        return False

    if r == {}:
        return False

    return r

def expr_simp_inf_signed(expr_simp, e):
    "((x - y) ^ ((x ^ y) & ((x - y) ^ x))) [31:32] == x <s y"

    arg = __check_msb(e)
    if arg is False:
        return e
    # We want jok3 = jok1 - jok2
    to_match = jok3 ^ ((jok1 ^ jok2) & (jok3 ^ jok1))
    r = __match_expr_wrap(arg,
                        to_match,
                        [jok1, jok2, jok3])

    if r is False:
        return e

    new_j3 = expr_simp(r[jok3])
    sub = expr_simp(r[jok1] - r[jok2])

    if new_j3 == sub:
        return ExprOp_inf_signed(r[jok1], r[jok2])
    else:
        return e

def expr_simp_inf_unsigned_inversed(expr_simp, e):
    "((x - y) ^ ((x ^ y) & ((x - y) ^ x))) ^ x ^ y [31:32] == x <u y"

    arg = __check_msb(e)
    if arg is False:
        return e

    # We want jok3 = jok1 - jok2
    to_match = jok3 ^ ((jok1 ^ jok2) & (jok3 ^ jok1)) ^ jok1 ^ jok2
    r = __match_expr_wrap(arg,
                        to_match,
                        [jok1, jok2, jok3])

    if r is False:
        return e

    new_j3 = expr_simp(r[jok3])
    sub = expr_simp(r[jok1] - r[jok2])

    if new_j3 == sub:
        return ExprOp_inf_unsigned(r[jok1], r[jok2])
    else:
        return e

def expr_simp_inverse(expr_simp, e):
    """(x <u y) ^ ((x ^ y) [31:32]) == x <s y,
    (x <s y) ^ ((x ^ y) [31:32]) == x <u y"""

    to_match = (ExprOp_inf_unsigned(jok1, jok2) ^ jok_small)
    r = __match_expr_wrap(e,
                        to_match,
                        [jok1, jok2, jok_small])

    # Check for 2 symmetric cases
    if r is False:
        to_match = (ExprOp_inf_signed(jok1, jok2) ^ jok_small)
        r = __match_expr_wrap(e,
                            to_match,
                            [jok1, jok2, jok_small])

        if r is False:
            return e
        cur_sig = m2_expr.TOK_INF_SIGNED
    else:
        cur_sig = m2_expr.TOK_INF_UNSIGNED


    arg = __check_msb(r[jok_small])
    if arg is False:
        return e

    if not isinstance(arg, m2_expr.ExprOp) or arg.op != "^":
        return e

    op_args = arg.args
    if len(op_args) != 2:
        return e

    if r[jok1] not in op_args or r[jok2] not in op_args:
        return e

    if cur_sig == m2_expr.TOK_INF_UNSIGNED:
        return ExprOp_inf_signed(r[jok1], r[jok2])
    else:
        return ExprOp_inf_unsigned(r[jok1], r[jok2])

def expr_simp_equal(expr_simp, e):
    """(x - y)?(0:1) == (x == y)"""

    to_match = m2_expr.ExprCond(jok1 + jok2, m2_expr.ExprInt(0, 1), m2_expr.ExprInt(1, 1))
    r = __match_expr_wrap(e,
                          to_match,
                          [jok1, jok2])
    if r is False:
        return e

    return ExprOp_equal(r[jok1], expr_simp(-r[jok2]))



#-*- coding:utf-8 -*-

from miasm.expression.expression import ExprAssign, ExprOp
from miasm.ir.ir import IRBlock, AssignBlock
from miasm.ir.analysis import LifterModelCall
from miasm.arch.mips32.sem import Lifter_Mips32l, Lifter_Mips32b

class LifterModelCallMips32l(Lifter_Mips32l, LifterModelCall):
    def __init__(self, loc_db):
        Lifter_Mips32l.__init__(self, loc_db)
        self.ret_reg = self.arch.regs.V0

    def call_effects(self, ad, instr):
        call_assignblk = AssignBlock(
            [
                ExprAssign(
                    self.ret_reg,
                    ExprOp(
                        'call_func_ret',
                        ad,
                        self.arch.regs.A0,
                        self.arch.regs.A1,
                        self.arch.regs.A2,
                        self.arch.regs.A3,
                    )
                ),
            ],
            instr
        )

        return [call_assignblk], []


    def add_asmblock_to_ircfg(self, block, ircfg, gen_pc_updt=False):
        """
        Add a native block to the current IR
        @block: native assembly block
        @ircfg: IRCFG instance
        @gen_pc_updt: insert PC update effects between instructions
        """
        loc_key = block.loc_key
        ir_blocks_all = []

        assignments = []
        for index, instr in enumerate(block.lines):
            if loc_key is None:
                assignments = []
                loc_key = self.get_loc_key_for_instr(instr)
            if instr.is_subcall():
                assert index == len(block.lines) - 2

                # Add last instruction first (before call)
                split = self.add_instr_to_current_state(
                    block.lines[-1], block, assignments,
                    ir_blocks_all, gen_pc_updt
                )
                assert not split
                # Add call effects after the delay splot
                split = self.add_instr_to_current_state(
                    instr, block, assignments,
                    ir_blocks_all, gen_pc_updt
                )
                assert split
                break
            split = self.add_instr_to_current_state(
                instr, block, assignments,
                ir_blocks_all, gen_pc_updt
            )
            if split:
                ir_blocks_all.append(IRBlock(self.loc_db, loc_key, assignments))
                loc_key = None
                assignments = []
        if loc_key is not None:
            ir_blocks_all.append(IRBlock(self.loc_db, loc_key, assignments))

        new_ir_blocks_all = self.post_add_asmblock_to_ircfg(block, ircfg, ir_blocks_all)
        for irblock in new_ir_blocks_all:
            ircfg.add_irblock(irblock)
        return new_ir_blocks_all

    def get_out_regs(self, _):
        return set([self.ret_reg, self.sp])

    def sizeof_char(self):
        return 8

    def sizeof_short(self):
        return 16

    def sizeof_int(self):
        return 32

    def sizeof_long(self):
        return 32

    def sizeof_pointer(self):
        return 32



class LifterModelCallMips32b(Lifter_Mips32b, LifterModelCallMips32l):
    def __init__(self, loc_db):
        Lifter_Mips32b.__init__(self, loc_db)
        self.ret_reg = self.arch.regs.V0



from __future__ import print_function
#
# Copyright (C) 2011 EADS France, Fabrice Desclaux <fabrice.desclaux@eads.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from hashlib import md5
import warnings

from future.utils import viewvalues

from miasm.core.asmblock import disasmEngine, AsmBlockBad
from miasm.core.interval import interval
from miasm.core.utils import BoundedDict
from miasm.expression.expression import LocKey
from miasm.jitter.csts import *

class JitCore(object):

    "JiT management. This is an abstract class"

    # Jitted function's name
    FUNCNAME = "block_entry"

    jitted_block_delete_cb = None
    jitted_block_max_size = 10000

    def __init__(self, lifter, bin_stream):
        """Initialise a JitCore instance.
        @lifter: Lifter instance for current architecture
        @bin_stream: bin_stream instance
        """
        # Arch related
        self.lifter = lifter
        self.ircfg = self.lifter.new_ircfg()
        self.arch_name = "%s%s" % (self.lifter.arch.name, self.lifter.attrib)

        # Structures for block tracking
        self.offset_to_jitted_func = BoundedDict(self.jitted_block_max_size,
                                       delete_cb=self.jitted_block_delete_cb)
        self.loc_key_to_block = {}
        self.blocks_mem_interval = interval()

        # Logging & options
        self.log_mn = False
        self.log_regs = False
        self.log_newbloc = False
        self.options = {"jit_maxline": 50,  # Maximum number of line jitted
                        "max_exec_per_call": 0 # 0 means no limit
                        }

        # Disassembly Engine
        self.split_dis = set()
        self.mdis = disasmEngine(
            lifter.arch, lifter.attrib, bin_stream,
            lines_wd=self.options["jit_maxline"],
            loc_db=lifter.loc_db,
            follow_call=False,
            dontdis_retcall=False,
            split_dis=self.split_dis,
        )

    @property
    def ir_arch(self):
        warnings.warn('DEPRECATION WARNING: use ".lifter" instead of ".ir_arch"')
        return self.lifter


    def set_options(self, **kwargs):
        "Set options relative to the backend"
        self.options.update(kwargs)

    def clear_jitted_blocks(self):
        "Reset all jitted blocks"
        self.offset_to_jitted_func.clear()
        self.loc_key_to_block.clear()
        self.blocks_mem_interval = interval()

    def add_disassembly_splits(self, *args):
        """The disassembly engine will stop on address in args if they
        are not at the block beginning"""
        self.split_dis.update(set(args))

    def remove_disassembly_splits(self, *args):
        """The disassembly engine will no longer stop on address in args"""
        self.split_dis.difference_update(set(args))

    def load(self):
        "Initialise the Jitter"
        raise NotImplementedError("Abstract class")

    def set_block_min_max(self, cur_block):
        "Update cur_block to set min/max address"

        if cur_block.lines:
            cur_block.ad_min = cur_block.lines[0].offset
            cur_block.ad_max = cur_block.lines[-1].offset + cur_block.lines[-1].l
        else:
            # 1 byte block for unknown mnemonic
            offset = self.lifter.loc_db.get_location_offset(cur_block.loc_key)
            cur_block.ad_min = offset
            cur_block.ad_max = offset+1


    def add_block_to_mem_interval(self, vm, block):
        "Update vm to include block addresses in its memory range"
        self.blocks_mem_interval += interval([(block.ad_min, block.ad_max - 1)])

        vm.reset_code_bloc_pool()
        for a, b in self.blocks_mem_interval:
            vm.add_code_bloc(a, b + 1)

    def add_block(self, block):
        """Add a block to JiT and JiT it.
        @block: asm_bloc to add
        """
        raise NotImplementedError("Abstract class")

    def disasm_and_jit_block(self, addr, vm):
        """Disassemble a new block and JiT it
        @addr: address of the block to disassemble (LocKey or int)
        @vm: VmMngr instance
        """

        # Get the block
        if isinstance(addr, LocKey):
            addr = self.lifter.loc_db.get_location_offset(addr)
            if addr is None:
                raise RuntimeError("Unknown offset for LocKey")

        # Prepare disassembler
        self.mdis.lines_wd = self.options["jit_maxline"]

        # Disassemble it
        cur_block = self.mdis.dis_block(addr)
        if isinstance(cur_block, AsmBlockBad):
            return cur_block
        # Logging
        if self.log_newbloc:
            print(cur_block)

        # Update label -> block
        self.loc_key_to_block[cur_block.loc_key] = cur_block

        # Store min/max block address needed in jit automod code
        self.set_block_min_max(cur_block)

        # JiT it
        self.add_block(cur_block)

        # Update jitcode mem range
        self.add_block_to_mem_interval(vm, cur_block)
        return cur_block

    def run_at(self, cpu, offset, stop_offsets):
        """Run from the starting address @offset.
        Execution will stop if:
        - max_exec_per_call option is reached
        - a new, yet unknown, block is reached after the execution of block at
          address @offset
        - an address in @stop_offsets is reached
        @cpu: JitCpu instance
        @offset: starting address (int)
        @stop_offsets: set of address on which the jitter must stop
        """

        if offset is None:
            offset = getattr(cpu, self.lifter.pc.name)

        if offset not in self.offset_to_jitted_func:
            # Need to JiT the block
            cur_block = self.disasm_and_jit_block(offset, cpu.vmmngr)
            if isinstance(cur_block, AsmBlockBad):
                errno = cur_block.errno
                if errno == AsmBlockBad.ERROR_IO:
                    cpu.vmmngr.set_exception(EXCEPT_ACCESS_VIOL)
                elif errno == AsmBlockBad.ERROR_CANNOT_DISASM:
                    cpu.set_exception(EXCEPT_UNK_MNEMO)
                else:
                    raise RuntimeError("Unhandled disasm result %r" % errno)
                return offset

        # Run the block and update cpu/vmmngr state
        return self.exec_wrapper(offset, cpu, self.offset_to_jitted_func.data,
                                 stop_offsets,
                                 self.options["max_exec_per_call"])

    def blocks_to_memrange(self, blocks):
        """Return an interval instance standing for blocks addresses
        @blocks: list of AsmBlock instances
        """

        mem_range = interval()
        mem_range = interval([(block.ad_min, block.ad_max - 1) for block in blocks])
        return mem_range

    def __updt_jitcode_mem_range(self, vm):
        """Rebuild the VM blocks address memory range
        @vm: VmMngr instance
        """

        # Reset the current pool
        vm.reset_code_bloc_pool()

        # Add blocks in the pool
        for start, stop in self.blocks_mem_interval:
            vm.add_code_bloc(start, stop + 1)

    def del_block_in_range(self, ad1, ad2):
        """Find and remove jitted block in range [ad1, ad2].
        Return the list of block removed.
        @ad1: First address
        @ad2: Last address
        """

        # Find concerned blocks
        modified_blocks = set()
        for block in viewvalues(self.loc_key_to_block):
            if not block.lines:
                continue
            if block.ad_max <= ad1 or block.ad_min >= ad2:
                # Block not modified
                pass
            else:
                # Modified blocks
                modified_blocks.add(block)

        # Remove modified blocks
        for block in modified_blocks:
            try:
                for irblock in block.blocks:
                    # Remove offset -> jitted block link
                    offset = self.lifter.loc_db.get_location_offset(irblock.loc_key)
                    if offset in self.offset_to_jitted_func:
                        del(self.offset_to_jitted_func[offset])

            except AttributeError:
                # The block has never been translated in IR
                offset = self.lifter.loc_db.get_location_offset(block.loc_key)
                if offset in self.offset_to_jitted_func:
                    del(self.offset_to_jitted_func[offset])

            # Remove label -> block link
            del(self.loc_key_to_block[block.loc_key])

        # Re generate blocks intervals
        self.blocks_mem_interval = self.blocks_to_memrange(self.loc_key_to_block.values())

        return modified_blocks

    def updt_automod_code_range(self, vm, mem_range):
        """Remove jitted code in range @mem_range
        @vm: VmMngr instance
        @mem_range: list of start/stop addresses
        """
        for addr_start, addr_stop in mem_range:
            self.del_block_in_range(addr_start, addr_stop)
        self.__updt_jitcode_mem_range(vm)
        vm.reset_memory_access()

    def updt_automod_code(self, vm):
        """Remove jitted code updated by memory write
        @vm: VmMngr instance
        """
        mem_range = []
        for addr_start, addr_stop in vm.get_memory_write():
            mem_range.append((addr_start, addr_stop))
        self.updt_automod_code_range(vm, mem_range)

    def hash_block(self, block):
        """
        Build a hash of the block @block
        @block: asmblock
        """
        block_raw = b"".join(line.b for line in block.lines)
        offset = self.lifter.loc_db.get_location_offset(block.loc_key)
        block_hash = md5(
            b"%X_%s_%s_%s_%s" % (
                offset,
                self.arch_name.encode(),
                b'\x01' if self.log_mn else b'\x00',
                b'\x01' if self.log_regs else b'\x00',
                block_raw
            )
        ).hexdigest()
        return block_hash

    @property
    def disasm_cb(self):
        warnings.warn("Deprecated API: use .mdis.dis_block_callback")
        return self.mdis.dis_block_callback

    @disasm_cb.setter
    def disasm_cb(self, value):
        warnings.warn("Deprecated API: use .mdis.dis_block_callback")
        self.mdis.dis_block_callback = value



from future.utils import viewitems, viewvalues

from argparse import ArgumentParser
from miasm.analysis.binary import Container
from miasm.analysis.machine import Machine
from miasm.jitter.llvmconvert import LLVMType, LLVMContext_IRCompilation, LLVMFunction_IRCompilation
from llvmlite import ir as llvm_ir
from miasm.expression.simplifications import expr_simp_high_to_explicit
from miasm.core.locationdb import LocationDB

parser = ArgumentParser("LLVM export example")
parser.add_argument("target", help="Target binary")
parser.add_argument("addr", help="Target address")
parser.add_argument("--architecture", "-a", help="Force architecture")
args = parser.parse_args()
loc_db = LocationDB()
# This part focus on obtaining an IRCFG to transform #
cont = Container.from_stream(open(args.target, 'rb'), loc_db)
machine = Machine(args.architecture if args.architecture else cont.arch)
lifter = machine.lifter(loc_db)
dis = machine.dis_engine(cont.bin_stream, loc_db=loc_db)
asmcfg = dis.dis_multiblock(int(args.addr, 0))
ircfg = lifter.new_ircfg_from_asmcfg(asmcfg)
ircfg.simplify(expr_simp_high_to_explicit)
######################################################

# Instantiate a context and the function to fill
context = LLVMContext_IRCompilation()
context.lifter = lifter

func = LLVMFunction_IRCompilation(context, name="test")
func.ret_type = llvm_ir.VoidType()
func.init_fc()

# Here, as an example, we arbitrarily represent registers with global
# variables. Locals allocas are used for the computation during the function,
# and is finally saved in the aforementioned global variable.

# In other words, for each registers:
# entry:
#     ...
#     %reg_val_in = load i32 @REG
#     %REG = alloca i32
#     store i32 %reg_val_in, i32* %REG
#     ...
# exit:
#     ...
#     %reg_val_out = load i32 %REG
#     store i32 %reg_val_out, i32* @REG
#     ...

all_regs = set()
for block in viewvalues(ircfg.blocks):
    for irs in block.assignblks:
        for dst, src in viewitems(irs.get_rw(mem_read=True)):
            elem = src.union(set([dst]))
            all_regs.update(
                x for x in elem
                if x.is_id()
            )

reg2glob = {}
for var in all_regs:
    # alloca reg = global reg
    data = context.mod.globals.get(str(var), None)
    if data is None:
        data = llvm_ir.GlobalVariable(context.mod,  LLVMType.IntType(var.size), name=str(var))
    data.initializer = LLVMType.IntType(var.size)(0)
    value = func.builder.load(data)
    func.local_vars_pointers[var.name] = func.builder.alloca(llvm_ir.IntType(var.size), name=var.name)
    func.builder.store(value, func.local_vars_pointers[var.name])
    reg2glob[var] = data

# IRCFG is imported, without the final "ret void"
func.from_ircfg(ircfg, append_ret=False)

# Finish the saving of registers (temporary version to global)
for reg, glob in viewitems(reg2glob):
    value = func.builder.load(func.local_vars_pointers[reg.name])
    func.builder.store(value, glob)

# Finish the function
func.builder.ret_void()

# Get it back
open("out.ll", "w").write(str(func))
# The optimized CFG can be seen with:
# $ opt -O2 -dot-cfg -S out.ll && xdot cfg.test.dot



from miasm.ir.translators.translator import Translator
from miasm.core.utils import size2mask
from miasm.expression.expression import ExprInt, ExprCond, ExprCompose, \
    TOK_EQUAL, \
    TOK_INF_SIGNED, TOK_INF_UNSIGNED, \
    TOK_INF_EQUAL_SIGNED, TOK_INF_EQUAL_UNSIGNED, \
    is_associative

def int_size_to_bn(value, size):
    if size < 32:
        int_str = "%.8x" % value
        size_nibble = 8
    else:
        # size must be multiple of 4
        size = ((size + 31) // 32) * 32
        size_nibble = size // 4
        fmt_str = "%%.%dx" % size_nibble
        int_str = fmt_str % value
    assert len(int_str) == size_nibble
    return int_str, size_nibble


TOK_CMP_TO_NATIVE_C = {
    TOK_EQUAL: "==",
    TOK_INF_SIGNED: "<",
    TOK_INF_UNSIGNED: "<",
    TOK_INF_EQUAL_SIGNED: "<=",
    TOK_INF_EQUAL_UNSIGNED: "<=",
}

TOK_CMP_TO_BIGNUM_C = {
    TOK_EQUAL: "equal",
    TOK_INF_SIGNED: "inf_signed",
    TOK_INF_UNSIGNED: "inf_unsigned",
    TOK_INF_EQUAL_SIGNED: "inf_equal_signed",
    TOK_INF_EQUAL_UNSIGNED: "inf_equal_unsigned",
}


def get_c_common_next_pow2(size):
    # For uncommon expression size, use at least uint8
    size = max(size, 8)
    next_power = 1
    while next_power < size:
        next_power <<= 1
    size = next_power
    return size


class TranslatorC(Translator):
    "Translate a Miasm expression to an equivalent C code"

    # Implemented language
    __LANG__ = "C"

    # Operations translation
    dct_shift = {'a>>': "right_arith",
                 '>>': "right_logic",
                 '<<': "left_logic",
                 }
    dct_rot = {'<<<': 'rot_left',
               '>>>': 'rot_right',
               }

    NATIVE_INT_MAX_SIZE = 64

    def __init__(self, loc_db=None, **kwargs):
        """Instance a C translator
        @loc_db: LocationDB instance
        """
        super(TranslatorC, self).__init__(**kwargs)
        # symbol pool
        self.loc_db = loc_db

    def _size2mask(self, size):
        """Return a C string corresponding to the size2mask operation, with support for
        @size <= 64"""
        assert size <= 64
        mask = size2mask(size)
        return "0x%x" % mask

    def from_ExprId(self, expr):
        return str(expr)

    def from_ExprInt(self, expr):
        if expr.size <= self.NATIVE_INT_MAX_SIZE:
            assert expr.size <= 64
            out = "0x%x" % int(expr)
            if expr.size == 64:
                out += "ULL"
            return out
        value, int_size = int_size_to_bn(int(expr), expr.size)
        return 'bignum_from_string("%s", %d)' % (value, int_size)

    def from_ExprLoc(self, expr):
        loc_key = expr.loc_key
        if self.loc_db is None:
            return str(loc_key)
        offset = self.loc_db.get_location_offset(loc_key)
        if offset is None:
            return str(loc_key)

        if expr.size <= self.NATIVE_INT_MAX_SIZE:
            return "0x%x" % offset

        value, int_size = int_size_to_bn(offset, 64)
        return 'bignum_from_string("%s", %d)' % (value, int_size)

    def from_ExprAssign(self, expr):
        new_dst = self.from_expr(expr.dst)
        new_src = self.from_expr(expr.src)
        return "%s = %s" % (new_dst, new_src)

    def from_ExprCond(self, expr):
        cond = self.from_expr(expr.cond)
        src1 = self.from_expr(expr.src1)
        src2 = self.from_expr(expr.src2)
        if not expr.cond.size <= self.NATIVE_INT_MAX_SIZE:
            cond = "(!bignum_is_zero(%s))" % cond
        out = "(%s?%s:%s)" % (cond, src1, src2)
        return out

    def from_ExprMem(self, expr):
        ptr = expr.ptr
        if ptr.size <= self.NATIVE_INT_MAX_SIZE:
            new_ptr = self.from_expr(ptr)
            if expr.size <= self.NATIVE_INT_MAX_SIZE:
                # Native ptr, Native Mem
                return "MEM_LOOKUP_%.2d(jitcpu, %s)" % (expr.size, new_ptr)
            else:
                # Native ptr, BN mem
                return "MEM_LOOKUP_INT_BN(jitcpu, %d, %s)" % (expr.size, new_ptr)
        # BN ptr
        new_ptr = self.from_expr(ptr)

        if expr.size <= self.NATIVE_INT_MAX_SIZE:
            # BN ptr, Native Mem
            return "MEM_LOOKUP_BN_INT(jitcpu, %d, %s)" % (expr.size, new_ptr)
        else:
            # BN ptr, BN mem
            return "MEM_LOOKUP_BN_BN(jitcpu, %d, %s)" % (expr.size, new_ptr)

    def from_ExprOp(self, expr):
        if len(expr.args) == 1:
            if expr.op == 'parity':
                arg = expr.args[0]
                out = self.from_expr(arg)
                if arg.size <= self.NATIVE_INT_MAX_SIZE:
                    out = "(%s&%s)" % (out, self._size2mask(arg.size))
                else:
                    out = 'bignum_mask(%s, 8)' % (out, 8)
                    out = 'bignum_to_uint64(%s)' % out
                out = 'parity(%s)' % out
                return out

            elif expr.op.startswith("zeroExt_"):
                arg = expr.args[0]
                if expr.size == arg.size:
                    return arg
                return self.from_expr(ExprCompose(arg, ExprInt(0, expr.size - arg.size)))

            elif expr.op.startswith("signExt_"):
                arg = expr.args[0]
                if expr.size == arg.size:
                    return arg
                add_size = expr.size - arg.size
                new_expr = ExprCompose(
                    arg,
                    ExprCond(
                        arg.msb(),
                        ExprInt(size2mask(add_size), add_size),
                        ExprInt(0, add_size)
                    )
                )
                return self.from_expr(new_expr)


            elif expr.op in ['cntleadzeros', 'cnttrailzeros']:
                arg = expr.args[0]
                out = self.from_expr(arg)
                if arg.size <= self.NATIVE_INT_MAX_SIZE:
                    out = "%s(0x%x, %s)" % (expr.op, expr.args[0].size, out)
                else:
                    out = "bignum_%s(%s, %d)" % (expr.op, out, arg.size)
                return out

            elif expr.op == '!':
                arg = expr.args[0]
                out = self.from_expr(arg)
                if expr.size <= self.NATIVE_INT_MAX_SIZE:
                    out = "(~ %s)&%s" % (out, self._size2mask(arg.size))
                else:
                    out = "bignum_not(%s)" % out
                    out = "bignum_mask(%s, expr.size)" % out
                return out

            elif expr.op in [
                    "ftan", "frndint", "f2xm1", "fsin", "fsqrt", "fabs", "fcos",
                    "fchs",
            ]:
                return "fpu_%s%d(%s)" % (
                    expr.op,
                    expr.size,
                    self.from_expr(expr.args[0]),
                )
            elif (expr.op.startswith("access_")    or
                  expr.op.startswith("load_")      or
                  expr.op.startswith("fxam_c")):
                arg = expr.args[0]
                out = self.from_expr(arg)
                out = "%s(%s)" % (expr.op, out)
                return out

            elif expr.op == "-":
                arg = expr.args[0]
                out = self.from_expr(arg)
                if arg.size <= self.NATIVE_INT_MAX_SIZE:
                    out = "(%s(%s))" % (expr.op, out)
                    out = "(%s&%s)" % (out, self._size2mask(arg.size))
                else:
                    out = "bignum_sub(bignum_from_uint64(0), %s)" % out
                    out = "bignum_mask(%s, %d)"% (out, expr.size)
                return out

            elif expr.op.startswith("fpround_"):
                return "%s_fp%d(%s)" % (
                    expr.op,
                    expr.size,
                    self.from_expr(expr.args[0]),
                )
            elif expr.op == "sint_to_fp":
                size = expr.size
                arg = expr.args[0]
                if size not in [32, 64]:
                    raise RuntimeError(
                        "Unsupported size for sint_to_fp: %r" % size
                    )
                return "%s_%d(%s)" % (expr.op, size, self.from_expr(arg))
            elif expr.op.startswith("fp_to_sint"):
                dest_size = expr.size
                arg_size = expr.args[0].size
                if (arg_size, dest_size) in [
                        (32, 32), (64, 64), (64, 32),
                ]:
                    func = "fp%d_to_sint%d" % (arg_size, dest_size)
                else:
                    raise RuntimeError(
                        "Unsupported size for fp_to_sint: %r to %r" % (
                            arg_size,
                            dest_size
                        ))
                return "%s(%s)" % (func, self.from_expr(expr.args[0]))
            elif expr.op.startswith("fpconvert_fp"):
                dest_size = expr.size
                arg_size = expr.args[0].size
                if (arg_size, dest_size) in [
                        (32, 64), (64, 32)
                ]:
                    func = "fp%d_to_fp%d" % (arg_size, dest_size)
                else:
                    raise RuntimeError(
                        "Unsupported size for fpconvert: %r to %r" % (arg_size,
                                                                      dest_size)
                    )
                return "%s(%s)" % (func, self.from_expr(expr.args[0]))
            else:
                raise NotImplementedError('Unknown op: %r' % expr.op)

        elif len(expr.args) == 2:
            if expr.op in self.dct_shift:
                arg0 = self.from_expr(expr.args[0])
                arg1 = self.from_expr(expr.args[1])
                if expr.size <= self.NATIVE_INT_MAX_SIZE:
                    out = 'SHIFT_%s(%d, %s, %s)' % (
                        self.dct_shift[expr.op].upper(),
                        expr.args[0].size,
                        arg0,
                        arg1
                    )
                else:
                    op = {
                        "<<": "lshift",
                        ">>": "rshift",
                        "a>>": "a_rshift"
                    }
                    out = "bignum_%s(%s, bignum_to_uint64(%s))" % (
                        op[expr.op], arg0, arg1
                    )
                    out = "bignum_mask(%s, %d)"% (out, expr.size)
                return out

            elif is_associative(expr):
                args = [self.from_expr(arg)
                        for arg in expr.args]
                if expr.size <= self.NATIVE_INT_MAX_SIZE:
                    out = (" %s " % expr.op).join(args)
                    out = "((%s)&%s)" % (out, self._size2mask(expr.size))
                else:
                    op_to_bn_func = {
                    "+": "add",
                    "*": "mul",
                    "|": "or",
                    "^": "xor",
                    "&": "and",
                    }
                    args = list(expr.args)
                    out = self.from_expr(args.pop())
                    while args:
                        out = 'bignum_mask(bignum_%s(%s, %s), %d)' % (
                            op_to_bn_func[expr.op],
                            out,
                            self.from_expr(args.pop()),
                            expr.size
                    )
                return out

            elif expr.op in ['-']:
                return '(((%s&%s) %s (%s&%s))&%s)' % (
                    self.from_expr(expr.args[0]),
                    self._size2mask(expr.args[0].size),
                    str(expr.op),
                    self.from_expr(expr.args[1]),
                    self._size2mask(expr.args[1].size),
                    self._size2mask(expr.args[0].size)
                )
            elif expr.op in self.dct_rot:
                arg0 = self.from_expr(expr.args[0])
                arg1 = self.from_expr(expr.args[1])
                if expr.size <= self.NATIVE_INT_MAX_SIZE:
                    out = '(%s(%s, %s, %s) &%s)' % (
                        self.dct_rot[expr.op],
                        expr.args[0].size,
                        arg0,
                        arg1,
                        self._size2mask(expr.args[0].size),
                    )
                else:
                    op = {
                        ">>>": "ror",
                        "<<<": "rol"
                    }
                    out = "bignum_%s(%s, %d, bignum_to_uint64(%s))" % (
                        op[expr.op], arg0, expr.size, arg1
                    )
                    out = "bignum_mask(%s, %d)"% (out, expr.size)
                return out

            elif expr.op == 'x86_cpuid':
                return "%s(%s, %s)" % (expr.op,
                                       self.from_expr(expr.args[0]),
                                       self.from_expr(expr.args[1]))
            elif expr.op.startswith("fcom"):
                arg0 = self.from_expr(expr.args[0])
                arg1 = self.from_expr(expr.args[1])
                if not expr.args[0].size <= self.NATIVE_INT_MAX_SIZE:
                    raise ValueError("Bad semantic: fpu do operations do not support such size")
                out = "fpu_%s(%s, %s)" % (expr.op, arg0, arg1)
                return out

            elif expr.op in ["fadd", "fsub", "fdiv", 'fmul', "fscale",
                             "fprem", "fyl2x", "fpatan"]:
                arg0 = self.from_expr(expr.args[0])
                arg1 = self.from_expr(expr.args[1])
                if not expr.args[0].size <= self.NATIVE_INT_MAX_SIZE:
                    raise ValueError("Bad semantic: fpu do operations do not support such size")
                out = "fpu_%s%d(%s, %s)" % (expr.op, expr.size, arg0, arg1)
                return out

            elif expr.op == "segm":
                return "segm2addr(jitcpu, %s, %s)" % (
                    self.from_expr(expr.args[0]),
                    self.from_expr(expr.args[1])
                )

            elif expr.op in ['udiv', 'umod']:
                arg0 = self.from_expr(expr.args[0])
                arg1 = self.from_expr(expr.args[1])

                if expr.size <= self.NATIVE_INT_MAX_SIZE:
                    out = '%s%d(%s, %s)' % (
                        expr.op,
                        expr.args[0].size,
                        arg0,
                        arg1
                    )
                else:
                    out = "bignum_%s(%s, %s)" % (
                        expr.op,
                        arg0,
                        arg1
                    )
                    out = "bignum_mask(%s, %d)"% (out, expr.size)
                return out



            elif expr.op in ['sdiv', 'smod']:
                arg0 = self.from_expr(expr.args[0])
                arg1 = self.from_expr(expr.args[1])

                if expr.size <= self.NATIVE_INT_MAX_SIZE:
                    out = '%s%d(%s, %s)' % (
                        expr.op,
                        expr.args[0].size,
                        arg0,
                        arg1
                    )
                else:
                    out = "bignum_%s(%s, %s, %d)" % (
                        expr.op,
                        arg0,
                        arg1,
                        expr.size
                    )
                    out = "bignum_mask(%s, %d)"% (out, expr.size)
                return out

            elif expr.op in ["bcdadd", "bcdadd_cf"]:
                return "%s_%d(%s, %s)" % (
                    expr.op, expr.args[0].size,
                    self.from_expr(expr.args[0]),
                    self.from_expr(expr.args[1])
                )


            elif expr.op in [
                    TOK_EQUAL,
                    TOK_INF_SIGNED,
                    TOK_INF_UNSIGNED,
                    TOK_INF_EQUAL_SIGNED,
                    TOK_INF_EQUAL_UNSIGNED,
            ]:
                arg0, arg1 = expr.args
                if expr.size <= self.NATIVE_INT_MAX_SIZE:
                    size = get_c_common_next_pow2(arg0.size)
                    op = TOK_CMP_TO_NATIVE_C[expr.op]
                    if expr.op in [TOK_INF_SIGNED, TOK_INF_EQUAL_SIGNED]:
                        arg0 = arg0.signExtend(size)
                        arg1 = arg1.signExtend(size)
                        arg0_C = self.from_expr(arg0)
                        arg1_C = self.from_expr(arg1)
                        cast = "(int%d_t)" % size
                    else:
                        arg0 = arg0.signExtend(size)
                        arg1 = arg1.signExtend(size)
                        arg0_C = self.from_expr(arg0)
                        arg1_C = self.from_expr(arg1)
                        cast = "(uint%d_t)" % size
                    out = '((%s%s %s %s%s)?1:0)' % (
                        cast,
                        arg0_C,
                        op,
                        cast,
                        arg1_C
                    )
                else:
                    op = TOK_CMP_TO_BIGNUM_C[expr.op]
                    out = "bignum_is_%s(%s, %s)" % (
                        op,
                        arg0,
                        arg1
                    )
                    out = "bignum_mask(%s, %d)"% (out, expr.size)
                return out


            else:
                raise NotImplementedError('Unknown op: %r' % expr.op)

        elif len(expr.args) >= 3 and is_associative(expr):  # ?????
            oper = ['(%s&%s)' % (
                self.from_expr(arg),
                self._size2mask(arg.size),
            )
                    for arg in expr.args]
            oper = str(expr.op).join(oper)
            return "((%s)&%s)" % (
                oper,
                self._size2mask(expr.args[0].size)
            )
        else:
            raise NotImplementedError('Unknown op: %s' % expr.op)

    def from_ExprSlice(self, expr):
        out = self.from_expr(expr.arg)
        if expr.arg.size <= self.NATIVE_INT_MAX_SIZE:
            # XXX check mask for 64 bit & 32 bit compat
            out = "((%s>>%d) &%s)" % (
                out, expr.start,
                self._size2mask(expr.stop - expr.start)
            )
        else:
            out = "bignum_rshift(%s, %d)" % (out, expr.start)
            out = "bignum_mask(%s, %d)" % (out, expr.stop - expr.start)

            if expr.size <= self.NATIVE_INT_MAX_SIZE:
                # Convert bignum to int
                out = "bignum_to_uint64(%s)" % out
        return out

    def from_ExprCompose(self, expr):
        if expr.size <= self.NATIVE_INT_MAX_SIZE:

            out = []
            size = get_c_common_next_pow2(expr.size)
            dst_cast = "uint%d_t" % size
            for index, arg in expr.iter_args():
                out.append("(((%s)(%s & %s)) << %d)" % (
                    dst_cast,
                    self.from_expr(arg),
                    self._size2mask(arg.size),
                    index)
                )
            out = ' | '.join(out)
            return '(' + out + ')'
        else:
            # Convert all parts to bignum
            args = []
            for index, arg in expr.iter_args():
                arg_str = self.from_expr(arg)
                if arg.size <= self.NATIVE_INT_MAX_SIZE:
                    arg_str = '((%s) & %s)' % (arg_str, self._size2mask(arg.size))
                    arg_str = 'bignum_from_uint64(%s)' % arg_str
                else:
                    arg_str = 'bignum_mask(%s, %d)' % (arg_str, arg.size)
                arg_str = 'bignum_lshift(%s, %d)' % (arg_str, index)
                args.append(arg_str)
            out = args.pop()
            while args:
                arg = args.pop()
                out = "bignum_or(%s, %s)" % (out, arg)
            return out


# Register the class
Translator.register(TranslatorC)
